# -*- coding: utf-8 -*-
#
# Copyright (C) 2008-2022 EDF R&D
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
#
# See http://www.salome-platform.org/ or email : webmaster.salome@opencascade.com
#
# Author: Jean-Philippe Argaud, jean-philippe.argaud@edf.fr, EDF R&D

"""
    Gestion simple de rapports structurés créés par l'utilisateur.
"""
__author__ = "Jean-Philippe ARGAUD"
__all__ = []

import os.path

# ==============================================================================
# Classes de services non utilisateur

class _ReportPartM__(object):
    """
    Store and retrieve the data for C: internal class
    """
    def __init__(self, part="default"):
        self.__part    = str(part)
        self.__styles  = []
        self.__content = []

    def append(self, content, style="p", position=-1):
        if position == -1:
            self.__styles.append(style)
            self.__content.append(content)
        else:
            self.__styles.insert(position, style)
            self.__content.insert(position, content)
        return 0

    def get_styles(self):
        return self.__styles

    def get_content(self):
        return self.__content

class _ReportM__(object):
    """
    Store and retrieve the data for C: internal class
    """
    def __init__(self, part='default'):
        self.__document = {}
        self.__document[part] = _ReportPartM__(part)

    def append(self,  content, style="p", position=-1, part='default'):
        if part not in self.__document:
            self.__document[part] = _ReportPartM__(part)
        self.__document[part].append(content, style, position)
        return 0

    def get_styles(self):
        op = list(self.__document.keys()) ; op.sort()
        return [self.__document[k].get_styles() for k in op]

    def get_content(self):
        op = list(self.__document.keys()) ; op.sort()
        return [self.__document[k].get_content() for k in op]

    def clear(self):
        self.__init__()

class __ReportC__(object):
    """
    Get user commands, update M and V: user intertace to create the report
    """
    m = _ReportM__()

    def append(self, content="", style="p", position=-1, part="default"):
        return self.m.append(content, style, position, part)

    def retrieve(self):
        st = self.m.get_styles()
        ct = self.m.get_content()
        return st, ct

    def clear(self):
        self.m.clear()

class __ReportV__(object):
    """
    Interact with user and C: template for reports
    """

    default_filename="report.txt"

    def __init__(self, c):
        self.c = c

    def save(self, filename=None):
        if filename is None:
            filename = self.default_filename
        _filename = os.path.abspath(filename)
        #
        h = self.get()
        fid = open(_filename, 'w')
        fid.write(h)
        fid.close()
        return filename, _filename

    def retrieve(self):
        return self.c.retrieve()

    def __str__(self):
        return self.get()

    def close(self):
        del self.c
        return 0

# ==============================================================================
# Classes d'interface utilisateur : ReportViewIn*, ReportStorage
# Tags de structure : (title, h1, h2, h3, p, uli, oli, <b></b>, <i></i>)

class ReportViewInHtml(__ReportV__):
    """
    Report in HTML
    """

    default_filename="report.html"
    tags = {
        "oli":"li",
        "uli":"li",
        }

    def get(self):
        st, ct = self.retrieve()
        inuLi, inoLi = False, False
        pg = "<html>\n<head>"
        pg += "\n<title>Report in HTML</title>"
        pg += "\n</head>\n<body>"
        for k,ps in enumerate(st):
            pc = ct[k]
            try:
                ii = ps.index("title")
                title = pc[ii]
                pg += "%s\n%s\n%s"%('<hr noshade><h1 align="center">',title,'</h1><hr noshade>')
            except Exception:
                pass
            for i,s in enumerate(ps):
                c = pc[i]
                if s == "uli" and not inuLi:
                    pg += "\n<ul>"
                    inuLi = True
                elif s == "oli" and not inoLi:
                    pg += "\n<ol>"
                    inoLi = True
                elif s != "uli" and inuLi:
                    pg += "\n</ul>"
                    inuLi = False
                elif s != "oli" and inoLi:
                    pg += "\n</ol>"
                    inoLi = False
                elif s == "title":
                    continue
                for t in self.tags:
                    if s == t: s = self.tags[t]
                pg += "\n<%s>%s</%s>"%(s,c,s)
        pg += "\n</body>\n</html>"
        return pg

class ReportViewInRst(__ReportV__):
    """
    Report in RST
    """

    default_filename="report.rst"
    tags = {
        "p":["\n\n",""],
        "uli":["\n  - ",""],
        "oli":["\n  #. ",""],
        }
    titles = {
        # "title":["=","="],
        "h1":["","-"],
        "h2":["","+"],
        "h3":["","*"],
        }
    translation = {
        "<b>":"**",
        "<i>":"*",
        "</b>":"**",
        "</i>":"*",
        }

    def get(self):
        st, ct = self.retrieve()
        inuLi, inoLi = False, False
        pg = ""
        for k,ps in enumerate(st):
            pc = ct[k]
            try:
                ii = ps.index("title")
                title = pc[ii]
                pg += "%s\n%s\n%s"%("="*80,title,"="*80)
            except Exception:
                pass
            for i,s in enumerate(ps):
                c = pc[i]
                if s == "uli" and not inuLi:
                    pg += "\n"
                    inuLi = True
                elif s == "oli" and not inoLi:
                    pg += "\n"
                    inoLi = True
                elif s != "uli" and inuLi:
                    pg += "\n"
                    inuLi = False
                elif s != "oli" and inoLi:
                    pg += "\n"
                    inoLi = False
                for t in self.translation:
                    c = c.replace(t,self.translation[t])
                if s in self.titles.keys():
                    pg += "\n%s\n%s\n%s"%(self.titles[s][0]*len(c),c,self.titles[s][1]*len(c))
                elif s in self.tags.keys():
                    pg += "%s%s%s"%(self.tags[s][0],c,self.tags[s][1])
            pg += "\n"
        return pg

class ReportViewInPlainTxt(__ReportV__):
    """
    Report in plain TXT
    """

    default_filename="report.txt"
    tags = {
        "p":["\n",""],
        "uli":["\n  - ",""],
        "oli":["\n  - ",""],
        }
    titles = {
        # "title":["=","="],
        "h1":["",""],
        "h2":["",""],
        "h3":["",""],
        }
    translation = {
        "<b>":"",
        "<i>":"",
        "</b>":"",
        "</i>":"",
        }

    def get(self):
        st, ct = self.retrieve()
        inuLi, inoLi = False, False
        pg = ""
        for k,ps in enumerate(st):
            pc = ct[k]
            try:
                ii = ps.index("title")
                title = pc[ii]
                pg += "%s\n%s\n%s"%("="*80,title,"="*80)
            except Exception:
                pass
            for i,s in enumerate(ps):
                c = pc[i]
                if s == "uli" and not inuLi:
                    inuLi = True
                elif s == "oli" and not inoLi:
                    inoLi = True
                elif s != "uli" and inuLi:
                    inuLi = False
                elif s != "oli" and inoLi:
                    inoLi = False
                for t in self.translation:
                    c = c.replace(t,self.translation[t])
                if s in self.titles.keys():
                    pg += "\n%s\n%s\n%s"%(self.titles[s][0]*len(c),c,self.titles[s][1]*len(c))
                elif s in self.tags.keys():
                    pg += "\n%s%s%s"%(self.tags[s][0],c,self.tags[s][1])
            pg += "\n"
        return pg


# Interface utilisateur de stockage des informations
ReportStorage = __ReportC__

# ==============================================================================
if __name__ == "__main__":
    print('\n AUTODIAGNOSTIC\n')



# -*- coding: utf-8 -*-
#
# Copyright (C) 2008-2022 EDF R&D
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
#
# See http://www.salome-platform.org/ or email : webmaster.salome@opencascade.com
#
# Author: Jean-Philippe Argaud, jean-philippe.argaud@edf.fr, EDF R&D

__doc__ = """
    Standard Kalman Filter
"""
__author__ = "Jean-Philippe ARGAUD"

import numpy
from daCore.PlatformInfo import PlatformInfo
mpr = PlatformInfo().MachinePrecision()

# ==============================================================================
def ecwstdkf(selfA, Xb, Y, U, HO, CM, R, B, __storeState = False):
    """
    Correction
    """
    if selfA._parameters["EstimationOf"] == "Parameters":
        selfA._parameters["StoreInternalVariables"] = True
    #
    # Initialisations
    # ---------------
    Hm = HO["Tangent"].asMatrix(Xb)
    Hm = Hm.reshape(Y.size,Xb.size) # ADAO & check shape
    Ha = HO["Adjoint"].asMatrix(Xb)
    Ha = Ha.reshape(Xb.size,Y.size) # ADAO & check shape
    #
    HXb = Hm @ Xb
    HXb = HXb.reshape((-1,1))
    if Y.size != HXb.size:
        raise ValueError("The size %i of observations Y and %i of observed calculation H(X) are different, they have to be identical."%(Y.size,HXb.size))
    if max(Y.shape) != max(HXb.shape):
        raise ValueError("The shapes %s of observations Y and %s of observed calculation H(X) are different, they have to be identical."%(Y.shape,HXb.shape))
    #
    if selfA._parameters["StoreInternalVariables"] or \
        selfA._toStore("CostFunctionJ")  or selfA._toStore("CostFunctionJAtCurrentOptimum") or \
        selfA._toStore("CostFunctionJb") or selfA._toStore("CostFunctionJbAtCurrentOptimum") or \
        selfA._toStore("CostFunctionJo") or selfA._toStore("CostFunctionJoAtCurrentOptimum") or \
        selfA._toStore("CurrentOptimum") or selfA._toStore("APosterioriCovariance") or \
        (Y.size >  Xb.size):
        if isinstance(B,numpy.ndarray):
            BI = numpy.linalg.inv(B)
        else:
            BI = B.getI()
        RI = R.getI()
    #
    Innovation  = Y - HXb
    if selfA._parameters["EstimationOf"] == "Parameters":
        if CM is not None and "Tangent" in CM and U is not None: # Attention : si Cm est aussi dans H, doublon !
            Cm = CM["Tangent"].asMatrix(Xb)
            Cm = Cm.reshape(Xb.size,U.size) # ADAO & check shape
            Innovation = Innovation - (Cm @ U).reshape((-1,1))
    #
    # Calcul de l'analyse
    # -------------------
    if Y.size <= Xb.size:
        _HNHt = numpy.dot(Hm, B @ Ha)
        _A = R + _HNHt
        _u = numpy.linalg.solve( _A , Innovation )
        Xa = Xb + (B @ (Ha @ _u)).reshape((-1,1))
        K  = B @ (Ha @ numpy.linalg.inv(_A))
    else:
        _HtRH = numpy.dot(Ha, RI @ Hm)
        _A = BI + _HtRH
        _u = numpy.linalg.solve( _A , numpy.dot(Ha, RI @ Innovation) )
        Xa = Xb + _u.reshape((-1,1))
        K = numpy.linalg.inv(_A) @ (Ha @ RI.asfullmatrix(Y.size))
    #
    Pa = B - K @ (Hm @ B)
    Pa = (Pa + Pa.T) * 0.5 # Symétrie
    Pa = Pa + mpr*numpy.trace( Pa ) * numpy.identity(Xa.size) # Positivité
    #
    if __storeState: selfA._setInternalState("Xn", Xa)
    if __storeState: selfA._setInternalState("Pn", Pa)
    #--------------------------
    #
    selfA.StoredVariables["Analysis"].store( Xa )
    if selfA._toStore("SimulatedObservationAtCurrentAnalysis"):
        selfA.StoredVariables["SimulatedObservationAtCurrentAnalysis"].store( Hm @ Xa )
    if selfA._toStore("InnovationAtCurrentAnalysis"):
        selfA.StoredVariables["InnovationAtCurrentAnalysis"].store( Innovation )
    # ---> avec current state
    if selfA._parameters["StoreInternalVariables"] \
        or selfA._toStore("CurrentState"):
        selfA.StoredVariables["CurrentState"].store( Xa )
    if selfA._toStore("BMA"):
        selfA.StoredVariables["BMA"].store( numpy.ravel(Xb) - numpy.ravel(Xa) )
    if selfA._toStore("InnovationAtCurrentState"):
        selfA.StoredVariables["InnovationAtCurrentState"].store( Innovation )
    if selfA._toStore("SimulatedObservationAtCurrentState") \
        or selfA._toStore("SimulatedObservationAtCurrentOptimum"):
        selfA.StoredVariables["SimulatedObservationAtCurrentState"].store( HXb )
    # ---> autres
    if selfA._parameters["StoreInternalVariables"] \
        or selfA._toStore("CostFunctionJ") \
        or selfA._toStore("CostFunctionJb") \
        or selfA._toStore("CostFunctionJo") \
        or selfA._toStore("CurrentOptimum") or selfA._toStore("APosterioriCovariance"):
        Jb  = float( 0.5 * (Xa - Xb).T @ (BI @ (Xa - Xb)) )
        Jo  = float( 0.5 * Innovation.T @ (RI @ Innovation) )
        J   = Jb + Jo
        selfA.StoredVariables["CostFunctionJb"].store( Jb )
        selfA.StoredVariables["CostFunctionJo"].store( Jo )
        selfA.StoredVariables["CostFunctionJ" ].store( J )
        #
        if selfA._toStore("IndexOfOptimum") \
            or selfA._toStore("CurrentOptimum") \
            or selfA._toStore("CostFunctionJAtCurrentOptimum") \
            or selfA._toStore("CostFunctionJbAtCurrentOptimum") \
            or selfA._toStore("CostFunctionJoAtCurrentOptimum") \
            or selfA._toStore("SimulatedObservationAtCurrentOptimum"):
            IndexMin = numpy.argmin( selfA.StoredVariables["CostFunctionJ"][:] )
        if selfA._toStore("IndexOfOptimum"):
            selfA.StoredVariables["IndexOfOptimum"].store( IndexMin )
        if selfA._toStore("CurrentOptimum"):
            selfA.StoredVariables["CurrentOptimum"].store( selfA.StoredVariables["Analysis"][IndexMin] )
        if selfA._toStore("SimulatedObservationAtCurrentOptimum"):
            selfA.StoredVariables["SimulatedObservationAtCurrentOptimum"].store( selfA.StoredVariables["SimulatedObservationAtCurrentAnalysis"][IndexMin] )
        if selfA._toStore("CostFunctionJbAtCurrentOptimum"):
            selfA.StoredVariables["CostFunctionJbAtCurrentOptimum"].store( selfA.StoredVariables["CostFunctionJb"][IndexMin] )
        if selfA._toStore("CostFunctionJoAtCurrentOptimum"):
            selfA.StoredVariables["CostFunctionJoAtCurrentOptimum"].store( selfA.StoredVariables["CostFunctionJo"][IndexMin] )
        if selfA._toStore("CostFunctionJAtCurrentOptimum"):
            selfA.StoredVariables["CostFunctionJAtCurrentOptimum" ].store( selfA.StoredVariables["CostFunctionJ" ][IndexMin] )
    if selfA._toStore("APosterioriCovariance"):
        selfA.StoredVariables["APosterioriCovariance"].store( Pa )
    #
    return 0

# ==============================================================================
if __name__ == "__main__":
    print('\n AUTODIAGNOSTIC\n')



# -*- coding: utf-8 -*-
#
# Copyright (C) 2008-2018 EDF R&D
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
#
# See http://www.salome-platform.org/ or email : webmaster.salome@opencascade.com
#
# Author: Jean-Philippe Argaud, jean-philippe.argaud@edf.fr, EDF R&D

# from distutils.core import setup
from setuptools import setup, find_packages

from adao.daCore.version import version

# read the contents of README file
from os import path
this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'README.txt')) as f:
    long_description = f.read()

setup(
    name = "adao",
    packages = [
        "adao",
        "adao/daAlgorithms",
        "adao/daAlgorithms/Atoms",
        "adao/daCore",
        "adao/daEficas",
        "adao/daEficasWrapper",
        "adao/daGUI",
        "adao/daGuiImpl",
        "adao/daNumerics",
        "adao/daUtils",
        "adao/daYacsIntegration",
        "adao/daYacsSchemaCreator",
        ],
    install_requires=['numpy','scipy'],
    python_requires='>=3.6',
    version = version,
    description = "A module for Data Assimilation and Optimization",
    author = "Jean-Philippe Argaud",
    author_email = "jean-philippe.argaud@edf.fr",
    url = "http://www.salome-platform.org/",
    license = "GNU Library or Lesser General Public License (LGPL)",
    # download_url = "",
    # packages=find_packages(),
    keywords = [
        "optimization", "data assimilation", "calibration", "interpolation",
        "inverse problem", "tunning", "minimization", "black-box", "checking",
        "3D-Var", "4D-Var", "Filtering", "Kalman", "Ensemble", "EnKF", "UKF",
        "BLUE", "Regression", "Quantile", "V&V", "Tabu Search",
        "DFO", "Derivative Free Optimization",
        "PSO", "Particle Swarm Optimization", "Swarm",
        "Gradient Test", "Adjoint Test",
        ],
    classifiers = [
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Development Status :: 5 - Production/Stable",
        "Environment :: Other Environment",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: GNU Library or Lesser General Public License (LGPL)",
        "Operating System :: OS Independent",
        "Topic :: Scientific/Engineering",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: Scientific/Engineering :: Mathematics",
        ],
    long_description = long_description,
    long_description_content_type='text/x-rst'
)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2008-2022 EDF R&D
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
#
# See http://www.salome-platform.org/ or email : webmaster.salome@opencascade.com
#
# Author: Jean-Philippe Argaud, jean-philippe.argaud@edf.fr, EDF R&D

__doc__ = """
    MMQR
"""
__author__ = "Jean-Philippe ARGAUD"

import sys, math, numpy
from daCore.PlatformInfo import PlatformInfo
mpr = PlatformInfo().MachinePrecision()
mfp = PlatformInfo().MaximumPrecision()

# ==============================================================================
def mmqr(
        func     = None,
        x0       = None,
        fprime   = None,
        bounds   = None,
        quantile = 0.5,
        maxfun   = 15000,
        toler    = 1.e-06,
        y        = None,
        ):
    """
    Implémentation informatique de l'algorithme MMQR, basée sur la publication :
    David R. Hunter, Kenneth Lange, "Quantile Regression via an MM Algorithm",
    Journal of Computational and Graphical Statistics, 9, 1, pp.60-77, 2000.
    """
    #
    # Recuperation des donnees et informations initiales
    # --------------------------------------------------
    variables = numpy.ravel( x0 )
    mesures   = numpy.ravel( y )
    increment = sys.float_info[0]
    p         = variables.size
    n         = mesures.size
    quantile  = float(quantile)
    #
    # Calcul des parametres du MM
    # ---------------------------
    tn      = float(toler) / n
    e0      = -tn / math.log(tn)
    epsilon = (e0-tn)/(1+math.log(e0))
    #
    # Calculs d'initialisation
    # ------------------------
    residus  = mesures - numpy.ravel( func( variables ) )
    poids    = 1./(epsilon+numpy.abs(residus))
    veps     = 1. - 2. * quantile - residus * poids
    lastsurrogate = -numpy.sum(residus*veps) - (1.-2.*quantile)*numpy.sum(residus)
    iteration = 0
    #
    # Recherche iterative
    # -------------------
    while (increment > toler) and (iteration < maxfun) :
        iteration += 1
        #
        Derivees  = numpy.array(fprime(variables))
        Derivees  = Derivees.reshape(n,p) # ADAO & check shape
        DeriveesT = Derivees.transpose()
        M         =   numpy.dot( DeriveesT , (numpy.array(p*[poids,]).T * Derivees) )
        SM        =   numpy.transpose(numpy.dot( DeriveesT , veps ))
        step      = - numpy.linalg.lstsq( M, SM, rcond=-1 )[0]
        #
        variables = variables + step
        if bounds is not None:
            # Attention : boucle infinie à éviter si un intervalle est trop petit
            while( (variables < numpy.ravel(numpy.asarray(bounds)[:,0])).any() or (variables > numpy.ravel(numpy.asarray(bounds)[:,1])).any() ):
                step      = step/2.
                variables = variables - step
        residus   = mesures - numpy.ravel( func(variables) )
        surrogate = numpy.sum(residus**2 * poids) + (4.*quantile-2.) * numpy.sum(residus)
        #
        while ( (surrogate > lastsurrogate) and ( max(list(numpy.abs(step))) > 1.e-16 ) ) :
            step      = step/2.
            variables = variables - step
            residus   = mesures - numpy.ravel( func(variables) )
            surrogate = numpy.sum(residus**2 * poids) + (4.*quantile-2.) * numpy.sum(residus)
        #
        increment     = lastsurrogate-surrogate
        poids         = 1./(epsilon+numpy.abs(residus))
        veps          = 1. - 2. * quantile - residus * poids
        lastsurrogate = -numpy.sum(residus * veps) - (1.-2.*quantile)*numpy.sum(residus)
    #
    # Mesure d'écart
    # --------------
    Ecart = quantile * numpy.sum(residus) - numpy.sum( residus[residus<0] )
    #
    return variables, Ecart, [n,p,iteration,increment,0]

# ==============================================================================
if __name__ == "__main__":
    print('\n AUTODIAGNOSTIC\n')



# -*- coding: utf-8 -*-
#
# Copyright (C) 2008-2022 EDF R&D
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
#
# See http://www.salome-platform.org/ or email : webmaster.salome@opencascade.com
#
# Author: Jean-Philippe Argaud, jean-philippe.argaud@edf.fr, EDF R&D

__doc__ = """
    Ensemble Kalman Smoother
"""
__author__ = "Jean-Philippe ARGAUD"

import copy, math, numpy, scipy, scipy.optimize, scipy.version
from daCore.NumericObjects import EnsembleErrorCovariance
from daCore.NumericObjects import EnsembleOfAnomalies
from daCore.NumericObjects import EnsembleOfBackgroundPerturbations
from daCore.NumericObjects import EnsemblePerturbationWithGivenCovariance
from daAlgorithms.Atoms import etkf
from daCore.PlatformInfo import PlatformInfo
mpr = PlatformInfo().MachinePrecision()
mfp = PlatformInfo().MaximumPrecision()

# ==============================================================================
def enks(selfA, Xb, Y, U, HO, EM, CM, R, B, Q, VariantM="EnKS16-KalmanFilterFormula"):
    """
    Ensemble Kalman Smoother
    """
    #
    # Opérateurs
    H = HO["Direct"].appliedControledFormTo
    #
    if selfA._parameters["EstimationOf"] == "State":
        M = EM["Direct"].appliedControledFormTo
    #
    if CM is not None and "Tangent" in CM and U is not None:
        Cm = CM["Tangent"].asMatrix(Xb)
    else:
        Cm = None
    #
    # Précalcul des inversions de B et R
    RIdemi = R.sqrtmI()
    #
    # Durée d'observation et tailles
    LagL = selfA._parameters["SmootherLagL"]
    if (not hasattr(Y,"store")) or (not hasattr(Y,"stepnumber")):
        raise ValueError("Fixed-lag smoother requires a series of observation")
    if Y.stepnumber() < LagL:
        raise ValueError("Fixed-lag smoother requires a series of observation greater then the lag L")
    duration = Y.stepnumber()
    __p = numpy.cumprod(Y.shape())[-1]
    __n = Xb.size
    __m = selfA._parameters["NumberOfMembers"]
    #
    if len(selfA.StoredVariables["Analysis"])==0 or not selfA._parameters["nextStep"]:
        selfA.StoredVariables["Analysis"].store( Xb )
        if selfA._toStore("APosterioriCovariance"):
            if hasattr(B,"asfullmatrix"):
                selfA.StoredVariables["APosterioriCovariance"].store( B.asfullmatrix(__n) )
            else:
                selfA.StoredVariables["APosterioriCovariance"].store( B )
    #
    # Calcul direct initial (on privilégie la mémorisation au recalcul)
    __seed = numpy.random.get_state()
    selfB = copy.deepcopy(selfA)
    selfB._parameters["StoreSupplementaryCalculations"] = ["CurrentEnsembleState"]
    if VariantM == "EnKS16-KalmanFilterFormula":
        etkf.etkf(selfB, Xb, Y, U, HO, EM, CM, R, B, Q, VariantM = "KalmanFilterFormula")
    else:
        raise ValueError("VariantM has to be chosen in the authorized methods list.")
    if LagL > 0:
        EL  = selfB.StoredVariables["CurrentEnsembleState"][LagL-1]
    else:
        EL = EnsembleOfBackgroundPerturbations( Xb, None, __m ) # Cf. etkf
    selfA._parameters["SetSeed"] = numpy.random.set_state(__seed)
    #
    for step in range(LagL,duration-1):
        #
        sEL = selfB.StoredVariables["CurrentEnsembleState"][step+1-LagL:step+1]
        sEL.append(None)
        #
        if hasattr(Y,"store"):
            Ynpu = numpy.ravel( Y[step+1] ).reshape((__p,1))
        else:
            Ynpu = numpy.ravel( Y ).reshape((__p,1))
        #
        if U is not None:
            if hasattr(U,"store") and len(U)>1:
                Un = numpy.ravel( U[step] ).reshape((-1,1))
            elif hasattr(U,"store") and len(U)==1:
                Un = numpy.ravel( U[0] ).reshape((-1,1))
            else:
                Un = numpy.ravel( U ).reshape((-1,1))
        else:
            Un = None
        #
        #--------------------------
        if VariantM == "EnKS16-KalmanFilterFormula":
            if selfA._parameters["EstimationOf"] == "State": # Forecast
                EL = M( [(EL[:,i], Un) for i in range(__m)],
                    argsAsSerie = True,
                    returnSerieAsArrayMatrix = True )
                EL = EnsemblePerturbationWithGivenCovariance( EL, Q )
                EZ = H( [(EL[:,i], Un) for i in range(__m)],
                    argsAsSerie = True,
                    returnSerieAsArrayMatrix = True )
                if Cm is not None and Un is not None: # Attention : si Cm est aussi dans M, doublon !
                    Cm = Cm.reshape(__n,Un.size) # ADAO & check shape
                    EZ = EZ + Cm @ Un
            elif selfA._parameters["EstimationOf"] == "Parameters":
                # --- > Par principe, M = Id, Q = 0
                EZ = H( [(EL[:,i], Un) for i in range(__m)],
                    argsAsSerie = True,
                    returnSerieAsArrayMatrix = True )
            #
            vEm   = EL.mean(axis=1, dtype=mfp).astype('float').reshape((__n,1))
            vZm   = EZ.mean(axis=1, dtype=mfp).astype('float').reshape((__p,1))
            #
            mS    = RIdemi @ EnsembleOfAnomalies( EZ, vZm, 1./math.sqrt(__m-1) )
            mS    = mS.reshape((-1,__m)) # Pour dimension 1
            delta = RIdemi @ ( Ynpu - vZm )
            mT    = numpy.linalg.inv( numpy.identity(__m) + mS.T @ mS )
            vw    = mT @ mS.T @ delta
            #
            Tdemi = numpy.real(scipy.linalg.sqrtm(mT))
            mU    = numpy.identity(__m)
            wTU   = (vw.reshape((__m,1)) + math.sqrt(__m-1) * Tdemi @ mU)
            #
            EX    = EnsembleOfAnomalies( EL, vEm, 1./math.sqrt(__m-1) )
            EL    = vEm + EX @ wTU
            #
            sEL[LagL] = EL
            for irl in range(LagL): # Lissage des L précédentes analysis
                vEm = sEL[irl].mean(axis=1, dtype=mfp).astype('float').reshape((__n,1))
                EX = EnsembleOfAnomalies( sEL[irl], vEm, 1./math.sqrt(__m-1) )
                sEL[irl] = vEm + EX @ wTU
            #
            # Conservation de l'analyse retrospective d'ordre 0 avant rotation
            Xa = sEL[0].mean(axis=1, dtype=mfp).astype('float').reshape((__n,1))
            if selfA._toStore("APosterioriCovariance"):
                EXn = sEL[0]
            #
            for irl in range(LagL):
                sEL[irl] = sEL[irl+1]
            sEL[LagL] = None
        #--------------------------
        else:
            raise ValueError("VariantM has to be chosen in the authorized methods list.")
        #
        selfA.StoredVariables["CurrentIterationNumber"].store( len(selfA.StoredVariables["Analysis"]) )
        # ---> avec analysis
        selfA.StoredVariables["Analysis"].store( Xa )
        if selfA._toStore("APosterioriCovariance"):
            selfA.StoredVariables["APosterioriCovariance"].store( EnsembleErrorCovariance(EXn) )
    #
    # Stockage des dernières analyses incomplètement remises à jour
    for irl in range(LagL):
        selfA.StoredVariables["CurrentIterationNumber"].store( len(selfA.StoredVariables["Analysis"]) )
        Xa = sEL[irl].mean(axis=1, dtype=mfp).astype('float').reshape((__n,1))
        selfA.StoredVariables["Analysis"].store( Xa )
    #
    return 0

# ==============================================================================
if __name__ == "__main__":
    print('\n AUTODIAGNOSTIC\n')



#-*- coding: utf-8 -*-
#
# Copyright (C) 2008-2022 EDF R&D
#
# This file is part of SALOME ADAO module
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
#
# See http://www.salome-platform.org/ or email : webmaster.salome@opencascade.com
#
# Author: Andre Ribes, andre.ribes@edf.fr, EDF R&D

import sys
import traceback
import logging
import pilot
import loader
import SALOMERuntime
import os

from daYacsSchemaCreator.infos_daComposant import *

def _Internal_Add_dir_script_ports(node, sf, ed, br, t_type):
  # On conserve le pointeur de "node" et "t_type"
  __scriptfile = str( sf )
  __exist_dir  = bool( ed )
  __base_dir   = str( br )
  __full_name  = os.path.join(__base_dir, os.path.basename(__scriptfile))
  if os.path.exists(__full_name):
    node.getInputPort("script").edInitPy(__full_name)
  else:
    node.getInputPort("script").edInitPy(__scriptfile)
  if __exist_dir:
    node.edAddInputPort("studydir", t_type)
    node.getInputPort("studydir").edInitPy(__base_dir)

def create_yacs_proc(study_config):

  logging.debug("[create_yacs_proc]")

  # Init part
  SALOMERuntime.RuntimeSALOME_setRuntime()
  l = loader.YACSLoader()
  l.registerProcCataLoader()
  runtime = pilot.getRuntime()
  try:
    catalogAd = runtime.loadCatalog("proc", os.environ["ADAO_ENGINE_ROOT_DIR"] + "/share/resources/adao/ADAOSchemaCatalog.xml")
    runtime.addCatalog(catalogAd)
  except:
    raise ValueError("Exception in loading ADAO YACS Schema catalog")

  # Starting creating proc
  proc = runtime.createProc("proc")
  proc.setTypeCode("pyobj", runtime.getTypeCode("pyobj"))
  proc.setTypeCode("SALOME_TYPES/ParametricInput", catalogAd._typeMap["SALOME_TYPES/ParametricInput"])
  proc.setTypeCode("SALOME_TYPES/ParametricOutput", catalogAd._typeMap["SALOME_TYPES/ParametricOutput"])
  t_pyobj  = proc.getTypeCode("pyobj")
  t_string = proc.getTypeCode("string")
  t_bool = proc.getTypeCode("bool")
  t_param_input  = proc.getTypeCode("SALOME_TYPES/ParametricInput")
  t_param_output = proc.getTypeCode("SALOME_TYPES/ParametricOutput")
  if "Repertory" in list(study_config.keys()):
    base_repertory = study_config["Repertory"]
    repertory      = True
  else:
    base_repertory = ""
    repertory = False
  if "ExecuteInContainer" in list(study_config.keys()):
    if study_config["ExecuteInContainer"] == "Mono":
      ExecuteInContainer = True
      ExecuteInModeMulti = False
    elif study_config["ExecuteInContainer"] == "Multi":
      ExecuteInContainer = True
      ExecuteInModeMulti = True
    else: # cas "No"
      ExecuteInContainer = False
      ExecuteInModeMulti = False
  else:
    ExecuteInContainer = False
    ExecuteInModeMulti = False
  if ExecuteInContainer and bool(UseSeparateContainer):
    mycontainer = proc.createContainer("AdaoContainer")
    if ExecuteInModeMulti:
      # type : multi pour creer un nouveau container pour chaque noeud
      # type : mono pour réutiliser le même container (defaut)
      mycontainer.setProperty("type","multi")

  # Create ADAO case bloc
  ADAO_Case = runtime.createBloc("ADAO_Case_Bloc")
  proc.edAddChild(ADAO_Case)

  # Step 0: create assimilation study object
  factory_CAS_node = catalogAd.getNodeFromNodeMap("CreateAssimilationStudy")
  CAS_node = factory_CAS_node.cloneNode("CreateAssimilationStudy")
  CAS_node.getInputPort("Name").edInitPy(study_config["Name"])
  CAS_node.getInputPort("Algorithm").edInitPy(study_config["Algorithm"])
  if "Debug" in study_config and study_config["Debug"] == "1":
    CAS_node.getInputPort("Debug").edInitPy(True)
  else:
    CAS_node.getInputPort("Debug").edInitPy(False)

  # Ajout des Variables
  InputVariablesNames = []
  InputVariablesSizes = []
  for var in study_config["InputVariables"]["Order"]:
    InputVariablesNames.append(var)
    InputVariablesSizes.append(int(study_config["InputVariables"][var]))
  CAS_node.getInputPort("InputVariablesNames").edInitPy(InputVariablesNames)
  CAS_node.getInputPort("InputVariablesSizes").edInitPy(InputVariablesSizes)
  OutputVariablesNames = []
  OutputVariablesSizes = []
  for var in study_config["OutputVariables"]["Order"]:
    OutputVariablesNames.append(var)
    OutputVariablesSizes.append(int(study_config["OutputVariables"][var]))
  CAS_node.getInputPort("OutputVariablesNames").edInitPy(OutputVariablesNames)
  CAS_node.getInputPort("OutputVariablesSizes").edInitPy(OutputVariablesSizes)

  ADAO_Case.edAddChild(CAS_node)
  if ExecuteInContainer and bool(UseSeparateContainer):
    CAS_node.setExecutionMode(SALOMERuntime.PythonNode.REMOTE_STR)
    CAS_node.setContainer(mycontainer)

  # Add an observer init node if an user defines some
  factory_init_observers_node = catalogAd.getNodeFromNodeMap("SetObserversNode")
  init_observers_node = factory_init_observers_node.cloneNode("SetObservers")
  if "Observers" in list(study_config.keys()):
    node_script = init_observers_node.getScript()
    node_script += "has_observers = True\n"
    node_script += "observers = " + str(study_config["Observers"]) + "\n"
    init_observers_node.setScript(node_script)
    ADAO_Case.edAddChild(init_observers_node)
    ADAO_Case.edAddDFLink(init_observers_node.getOutputPort("has_observers"), CAS_node.getInputPort("has_observers"))
    ADAO_Case.edAddDFLink(init_observers_node.getOutputPort("observers"), CAS_node.getInputPort("observers"))
  else:
    node_script = init_observers_node.getScript()
    node_script += "has_observers = False\n"
    node_script += "observers = \"\"\n"
    init_observers_node.setScript(node_script)
    ADAO_Case.edAddChild(init_observers_node)
    ADAO_Case.edAddDFLink(init_observers_node.getOutputPort("has_observers"), CAS_node.getInputPort("has_observers"))
    ADAO_Case.edAddDFLink(init_observers_node.getOutputPort("observers"), CAS_node.getInputPort("observers"))
  if ExecuteInContainer and bool(UseSeparateContainer):
    init_observers_node.setExecutionMode(SALOMERuntime.PythonNode.REMOTE_STR)
    init_observers_node.setContainer(mycontainer)

  # Step 0.5: Find if there is a user init node
  init_config = {}
  init_config["Target"] = []
  if "UserDataInit" in list(study_config.keys()):
    init_config = study_config["UserDataInit"]
    factory_init_node = catalogAd.getNodeFromNodeMap("UserDataInitFromScript")
    init_node = factory_init_node.cloneNode("UserDataInit")
    _Internal_Add_dir_script_ports( init_node, init_config["Data"], repertory, base_repertory, t_string)
    init_node_script = init_node.getScript()
    init_node_script += "# Import script and get data\n__import__(module_name)\nuser_script_module = sys.modules[module_name]\n\n"
    init_node_script += "init_data = user_script_module.init_data\n"
    init_node.setScript(init_node_script)
    ADAO_Case.edAddChild(init_node)
    if ExecuteInContainer and bool(UseSeparateContainer):
      init_node.setExecutionMode(SALOMERuntime.PythonNode.REMOTE_STR)
      init_node.setContainer(mycontainer)

  # Step 1: get input data from user configuration

  st_keys = sorted(list(study_config.keys()))
  for key in st_keys:
    ad_keys = sorted(AssimData)
    if key in ad_keys:
      data_config = study_config[key]

      key_type = key + "Type"
      key_stored = key + "Stored"

      if data_config["Type"] == "Dict" and data_config["From"] == "Script":
        # Create node
        factory_back_node = catalogAd.getNodeFromNodeMap("CreateDictFromScript")
        back_node = factory_back_node.cloneNode("Get" + key)
        _Internal_Add_dir_script_ports( back_node, data_config["Data"], repertory, base_repertory, t_string)
        back_node.edAddOutputPort(key, t_pyobj)
        ADAO_Case.edAddChild(back_node)
        if ExecuteInContainer and bool(UseSeparateContainer):
          back_node.setExecutionMode(SALOMERuntime.PythonNode.REMOTE_STR)
          back_node.setContainer(mycontainer)
        # Set content of the node
        back_node_script = back_node.getScript()
        if key in init_config["Target"]:
          # Connect node with InitUserData
          back_node_script += "__builtins__[\"init_data\"] = init_data\n"
          back_node.edAddInputPort("init_data", t_pyobj)
          ADAO_Case.edAddDFLink(init_node.getOutputPort("init_data"), back_node.getInputPort("init_data"))
        back_node_script += "# Import script and get data\n__import__(module_name)\nuser_script_module = sys.modules[module_name]\n\n"
        if key == "AlgorithmParameters":
            back_node_script += "if hasattr(user_script_module,'" + key + "'):\n  "
            back_node_script += key + " = user_script_module." + key + "\n"
            skey = "Parameters"
            back_node_script += "elif hasattr(user_script_module,'" + skey + "'):\n  "
            back_node_script += key + " = user_script_module." + skey + "\n"
        else:
            back_node_script += key + " = user_script_module." + key + "\n"
        back_node.setScript(back_node_script)
        # Connect node with CreateAssimilationStudy
        CAS_node.edAddInputPort(key, t_pyobj)
        ADAO_Case.edAddDFLink(back_node.getOutputPort(key), CAS_node.getInputPort(key))

      if data_config["Type"] == "Dict" and data_config["From"] == "String":
        # Create node
        factory_back_node = catalogAd.getNodeFromNodeMap("CreateDictFromString")
        back_node = factory_back_node.cloneNode("Get" + key)
        back_node.getInputPort("dict_in_string").edInitPy(data_config["Data"])
        back_node.edAddOutputPort(key, t_pyobj)
        ADAO_Case.edAddChild(back_node)
        if ExecuteInContainer and bool(UseSeparateContainer):
          back_node.setExecutionMode(SALOMERuntime.PythonNode.REMOTE_STR)
          back_node.setContainer(mycontainer)
        # Set content of the node
        back_node_script = back_node.getScript()
        if key in init_config["Target"]:
          # Connect node with InitUserData
          back_node_script += "__builtins__[\"init_data\"] = init_data\n"
          back_node.edAddInputPort("init_data", t_pyobj)
          ADAO_Case.edAddDFLink(init_node.getOutputPort("init_data"), back_node.getInputPort("init_data"))
        back_node_script += key + " = dict(dico)\n"
        back_node_script += "logging.debug(\"Dict is %ss\"%s%s)"%("%","%",key)
        back_node.setScript(back_node_script)
        # Connect node with CreateAssimilationStudy
        CAS_node.edAddInputPort(key, t_pyobj)
        ADAO_Case.edAddDFLink(back_node.getOutputPort(key), CAS_node.getInputPort(key))

      if data_config["Type"] == "Vector" and data_config["From"] == "String":
        # Create node
        factory_back_node = catalogAd.getNodeFromNodeMap("CreateNumpyVectorFromString")
        back_node = factory_back_node.cloneNode("Get" + key)
        back_node.getInputPort("vector_in_string").edInitPy(data_config["Data"])
        ADAO_Case.edAddChild(back_node)
        if ExecuteInContainer and bool(UseSeparateContainer):
          back_node.setExecutionMode(SALOMERuntime.PythonNode.REMOTE_STR)
          back_node.setContainer(mycontainer)
        # Set content of the node
        back_node_script = back_node.getScript()
        if "Stored" in data_config:
            back_node_script += "stored = " + str(data_config["Stored"]) + "\n"
        else:
            back_node_script += "stored = 0\n"
        if key in init_config["Target"]:
          # Connect node with InitUserData
          back_node_script += "__builtins__[\"init_data\"] = init_data\n"
          back_node.edAddInputPort("init_data", t_pyobj)
          ADAO_Case.edAddDFLink(init_node.getOutputPort("init_data"), back_node.getInputPort("init_data"))
        back_node.setScript(back_node_script)
        # Connect node with CreateAssimilationStudy
        CAS_node.edAddInputPort(key, t_pyobj)
        CAS_node.edAddInputPort(key_type, t_string)
        CAS_node.edAddInputPort(key_stored, t_bool)
        ADAO_Case.edAddDFLink(back_node.getOutputPort("vector"), CAS_node.getInputPort(key))
        ADAO_Case.edAddDFLink(back_node.getOutputPort("type"), CAS_node.getInputPort(key_type))
        ADAO_Case.edAddDFLink(back_node.getOutputPort("stored"), CAS_node.getInputPort(key_stored))

      if data_config["Type"] == "Vector" and data_config["From"] == "Script":
        # Create node
        factory_back_node = catalogAd.getNodeFromNodeMap("CreateNumpyVectorFromScript")
        back_node = factory_back_node.cloneNode("Get" + key)
        _Internal_Add_dir_script_ports( back_node, data_config["Data"], repertory, base_repertory, t_string)
        back_node.edAddOutputPort(key, t_pyobj)
        ADAO_Case.edAddChild(back_node)
        if ExecuteInContainer and bool(UseSeparateContainer):
          back_node.setExecutionMode(SALOMERuntime.PythonNode.REMOTE_STR)
          back_node.setContainer(mycontainer)
        # Set content of the node
        back_node_script = back_node.getScript()
        if key in init_config["Target"]:
          # Connect node with InitUserData
          back_node_script += "__builtins__[\"init_data\"] = init_data\n"
          back_node.edAddInputPort("init_data", t_pyobj)
          ADAO_Case.edAddDFLink(init_node.getOutputPort("init_data"), back_node.getInputPort("init_data"))
        back_node_script += "# Import script and get data\n__import__(module_name)\nuser_script_module = sys.modules[module_name]\n\n"
        back_node_script += key + " = user_script_module." + key + "\n"
        if "Stored" in data_config:
            back_node_script += "stored = " + str(data_config["Stored"]) + "\n"
        else:
            back_node_script += "stored = 0\n"
        back_node.setScript(back_node_script)
        # Connect node with CreateAssimilationStudy
        CAS_node.edAddInputPort(key, t_pyobj)
        CAS_node.edAddInputPort(key_type, t_string)
        CAS_node.edAddInputPort(key_stored, t_bool)
        ADAO_Case.edAddDFLink(back_node.getOutputPort(key), CAS_node.getInputPort(key))
        ADAO_Case.edAddDFLink(back_node.getOutputPort("type"), CAS_node.getInputPort(key_type))
        ADAO_Case.edAddDFLink(back_node.getOutputPort("stored"), CAS_node.getInputPort(key_stored))

      if data_config["Type"] == "Vector" and data_config["From"] == "DataFile":
        # Create node
        factory_back_node = catalogAd.getNodeFromNodeMap("CreateNumpyVectorFromDataFile")
        back_node = factory_back_node.cloneNode("Get" + key)
        back_node.getInputPort("script").edInitPy(data_config["Data"])
        back_node.getInputPort("columns").edInitPy((key,)) # On impose le concept, et le schéma YACS est ammendable
        if "ColMajor" in data_config:
            colmajor = bool(data_config["ColMajor"])
        else:
            colmajor = False
        back_node.getInputPort("colmajor").edInitPy(colmajor) # On impose le concept, et le schéma YACS est ammendable
        ADAO_Case.edAddChild(back_node)
        if ExecuteInContainer and bool(UseSeparateContainer):
          back_node.setExecutionMode(SALOMERuntime.PythonNode.REMOTE_STR)
          back_node.setContainer(mycontainer)
        # Set content of the node
        back_node_script = back_node.getScript()
        if "Stored" in data_config:
            back_node_script += "stored = " + str(data_config["Stored"]) + "\n"
        else:
            back_node_script += "stored = 0\n"
        if key in init_config["Target"]:
          # Connect node with InitUserData
          back_node_script += "__builtins__[\"init_data\"] = init_data\n"
          back_node.edAddInputPort("init_data", t_pyobj)
          ADAO_Case.edAddDFLink(init_node.getOutputPort("init_data"), back_node.getInputPort("init_data"))
        back_node.setScript(back_node_script)
        # Connect node with CreateAssimilationStudy
        CAS_node.edAddInputPort(key, t_pyobj)
        CAS_node.edAddInputPort(key_type, t_string)
        CAS_node.edAddInputPort(key_stored, t_bool)
        ADAO_Case.edAddDFLink(back_node.getOutputPort("vector"), CAS_node.getInputPort(key))
        ADAO_Case.edAddDFLink(back_node.getOutputPort("type"), CAS_node.getInputPort(key_type))
        ADAO_Case.edAddDFLink(back_node.getOutputPort("stored"), CAS_node.getInputPort(key_stored))

      if data_config["Type"] == "VectorSerie" and data_config["From"] == "String":
        # Create node
        factory_back_node = catalogAd.getNodeFromNodeMap("CreateNumpyVectorSerieFromString")
        back_node = factory_back_node.cloneNode("Get" + key)
        back_node.getInputPort("vector_in_string").edInitPy(data_config["Data"])
        ADAO_Case.edAddChild(back_node)
        if ExecuteInContainer and bool(UseSeparateContainer):
          back_node.setExecutionMode(SALOMERuntime.PythonNode.REMOTE_STR)
          back_node.setContainer(mycontainer)
        # Set content of the node
        back_node_script = back_node.getScript()
        if "Stored" in data_config:
            back_node_script += "stored = " + str(data_config["Stored"]) + "\n"
        else:
            back_node_script += "stored = 0\n"
        if key in init_config["Target"]:
          # Connect node with InitUserData
          back_node_script += "__builtins__[\"init_data\"] = init_data\n" + back_node_script
          back_node.edAddInputPort("init_data", t_pyobj)
          ADAO_Case.edAddDFLink(init_node.getOutputPort("init_data"), back_node.getInputPort("init_data"))
        back_node.setScript(back_node_script)
        # Connect node with CreateAssimilationStudy
        CAS_node.edAddInputPort(key, t_pyobj)
        CAS_node.edAddInputPort(key_type, t_string)
        CAS_node.edAddInputPort(key_stored, t_bool)
        ADAO_Case.edAddDFLink(back_node.getOutputPort("vector"), CAS_node.getInputPort(key))
        ADAO_Case.edAddDFLink(back_node.getOutputPort("type"), CAS_node.getInputPort(key_type))
        ADAO_Case.edAddDFLink(back_node.getOutputPort("stored"), CAS_node.getInputPort(key_stored))

      if data_config["Type"] == "VectorSerie" and data_config["From"] == "Script":
        # Create node
        factory_back_node = catalogAd.getNodeFromNodeMap("CreateNumpyVectorSerieFromScript")
        back_node = factory_back_node.cloneNode("Get" + key)
        _Internal_Add_dir_script_ports( back_node, data_config["Data"], repertory, base_repertory, t_string)
        back_node.edAddOutputPort(key, t_pyobj)
        ADAO_Case.edAddChild(back_node)
        if ExecuteInContainer and bool(UseSeparateContainer):
          back_node.setExecutionMode(SALOMERuntime.PythonNode.REMOTE_STR)
          back_node.setContainer(mycontainer)
        # Set content of the node
        back_node_script = back_node.getScript()
        if key in init_config["Target"]:
          # Connect node with InitUserData
          back_node_script += "__builtins__[\"init_data\"] = init_data\n"
          back_node.edAddInputPort("init_data", t_pyobj)
          ADAO_Case.edAddDFLink(init_node.getOutputPort("init_data"), back_node.getInputPort("init_data"))
        back_node_script += "# Import script and get data\n__import__(module_name)\nuser_script_module = sys.modules[module_name]\n\n"
        back_node_script += key + " = user_script_module." + key + "\n"
        if "Stored" in data_config:
            back_node_script += "stored = " + str(data_config["Stored"]) + "\n"
        else:
            back_node_script += "stored = 0\n"
        back_node.setScript(back_node_script)
        # Connect node with CreateAssimilationStudy
        CAS_node.edAddInputPort(key, t_pyobj)
        CAS_node.edAddInputPort(key_type, t_string)
        CAS_node.edAddInputPort(key_stored, t_bool)
        ADAO_Case.edAddDFLink(back_node.getOutputPort(key), CAS_node.getInputPort(key))
        ADAO_Case.edAddDFLink(back_node.getOutputPort("type"), CAS_node.getInputPort(key_type))
        ADAO_Case.edAddDFLink(back_node.getOutputPort("stored"), CAS_node.getInputPort(key_stored))

      if data_config["Type"] == "VectorSerie" and data_config["From"] == "DataFile":
        # Create node
        factory_back_node = catalogAd.getNodeFromNodeMap("CreateNumpyVectorSerieFromDataFile")
        back_node = factory_back_node.cloneNode("Get" + key)
        back_node.getInputPort("script").edInitPy(data_config["Data"])
        back_node.getInputPort("columns").edInitPy(()) # On impose aucun nom et le schéma YACS est ammendable
        if "ColMajor" in data_config:
            colmajor = bool(data_config["ColMajor"])
        else:
            colmajor = False
        back_node.getInputPort("colmajor").edInitPy(colmajor) # On impose le concept, et le schéma YACS est ammendable
        ADAO_Case.edAddChild(back_node)
        if ExecuteInContainer and bool(UseSeparateContainer):
          back_node.setExecutionMode(SALOMERuntime.PythonNode.REMOTE_STR)
          back_node.setContainer(mycontainer)
        # Set content of the node
        back_node_script = back_node.getScript()
        if "Stored" in data_config:
            back_node_script += "stored = " + str(data_config["Stored"]) + "\n"
        else:
            back_node_script += "stored = 0\n"
        if key in init_config["Target"]:
          # Connect node with InitUserData
          back_node_script += "__builtins__[\"init_data\"] = init_data\n"
          back_node.edAddInputPort("init_data", t_pyobj)
          ADAO_Case.edAddDFLink(init_node.getOutputPort("init_data"), back_node.getInputPort("init_data"))
        back_node.setScript(back_node_script)
        # Connect node with CreateAssimilationStudy
        CAS_node.edAddInputPort(key, t_pyobj)
        CAS_node.edAddInputPort(key_type, t_string)
        CAS_node.edAddInputPort(key_stored, t_bool)
        ADAO_Case.edAddDFLink(back_node.getOutputPort("vector"), CAS_node.getInputPort(key))
        ADAO_Case.edAddDFLink(back_node.getOutputPort("type"), CAS_node.getInputPort(key_type))
        ADAO_Case.edAddDFLink(back_node.getOutputPort("stored"), CAS_node.getInputPort(key_stored))

      if data_config["Type"] in ("Matrix", "ScalarSparseMatrix", "DiagonalSparseMatrix") and data_config["From"] == "String":
        # Create node
        factory_back_node = catalogAd.getNodeFromNodeMap("CreateNumpy%sFromString"%(data_config["Type"],))
        back_node = factory_back_node.cloneNode("Get" + key)
        back_node.getInputPort("matrix_in_string").edInitPy(data_config["Data"])
        ADAO_Case.edAddChild(back_node)
        if ExecuteInContainer and bool(UseSeparateContainer):
          back_node.setExecutionMode(SALOMERuntime.PythonNode.REMOTE_STR)
          back_node.setContainer(mycontainer)
        # Set content of the node
        back_node_script = back_node.getScript()
        if "Stored" in data_config:
            back_node_script += "stored = " + str(data_config["Stored"]) + "\n"
        else:
            back_node_script += "stored = 0\n"
        if key in init_config["Target"]:
          # Connect node with InitUserData
          back_node_script += "__builtins__[\"init_data\"] = init_data\n"
          back_node.edAddInputPort("init_data", t_pyobj)
          ADAO_Case.edAddDFLink(init_node.getOutputPort("init_data"), back_node.getInputPort("init_data"))
        back_node.setScript(back_node_script)
        # Connect node with CreateAssimilationStudy
        CAS_node.edAddInputPort(key, t_pyobj)
        CAS_node.edAddInputPort(key_type, t_string)
        CAS_node.edAddInputPort(key_stored, t_bool)
        ADAO_Case.edAddDFLink(back_node.getOutputPort("matrix"), CAS_node.getInputPort(key))
        ADAO_Case.edAddDFLink(back_node.getOutputPort("type"), CAS_node.getInputPort(key_type))
        ADAO_Case.edAddDFLink(back_node.getOutputPort("stored"), CAS_node.getInputPort(key_stored))

      if data_config["Type"] in ("Matrix", "ScalarSparseMatrix", "DiagonalSparseMatrix") and data_config["From"] == "Script":
        # Create node
        factory_back_node = catalogAd.getNodeFromNodeMap("CreateNumpy%sFromScript"%(data_config["Type"],))
        back_node = factory_back_node.cloneNode("Get" + key)
        _Internal_Add_dir_script_ports( back_node, data_config["Data"], repertory, base_repertory, t_string)
        back_node.edAddOutputPort(key, t_pyobj)
        ADAO_Case.edAddChild(back_node)
        if ExecuteInContainer and bool(UseSeparateContainer):
          back_node.setExecutionMode(SALOMERuntime.PythonNode.REMOTE_STR)
          back_node.setContainer(mycontainer)
        # Set content of the node
        back_node_script = back_node.getScript()
        if "Stored" in data_config:
            back_node_script += "stored = " + str(data_config["Stored"]) + "\n"
        else:
            back_node_script += "stored = 0\n"
        if key in init_config["Target"]:
          # Connect node with InitUserData
          back_node_script += "__builtins__[\"init_data\"] = init_data\n"
          back_node.edAddInputPort("init_data", t_pyobj)
          ADAO_Case.edAddDFLink(init_node.getOutputPort("init_data"), back_node.getInputPort("init_data"))
        back_node_script += "# Import script and get data\n__import__(module_name)\nuser_script_module = sys.modules[module_name]\n\n"
        back_node_script += key + " = user_script_module." + key + "\n"
        back_node.setScript(back_node_script)
        # Connect node with CreateAssimilationStudy
        CAS_node.edAddInputPort(key, t_pyobj)
        CAS_node.edAddInputPort(key_type, t_string)
        CAS_node.edAddInputPort(key_stored, t_bool)
        ADAO_Case.edAddDFLink(back_node.getOutputPort(key), CAS_node.getInputPort(key))
        ADAO_Case.edAddDFLink(back_node.getOutputPort("type"), CAS_node.getInputPort(key_type))
        ADAO_Case.edAddDFLink(back_node.getOutputPort("stored"), CAS_node.getInputPort(key_stored))

      if data_config["Type"] == "Function" and (key == "ObservationOperator" or key == "EvolutionModel"):
         TheData = data_config["Data"]
         for FunctionName in TheData["Function"]:
           port_name = key + FunctionName
           CAS_node.edAddInputPort(port_name, t_string)
           if os.path.exists(os.path.join(base_repertory, os.path.basename(TheData["Script"][FunctionName]))):
             CAS_node.getInputPort(port_name).edInitPy(os.path.join(base_repertory, os.path.basename(TheData["Script"][FunctionName])))
           else:
             CAS_node.getInputPort(port_name).edInitPy(TheData["Script"][FunctionName])
           try:
             CAS_node.edAddInputPort("studydir", t_string)
             CAS_node.getInputPort("studydir").edInitPy(base_repertory)
           except: pass

  # Step 3: create compute bloc
  compute_bloc = runtime.createBloc("compute_bloc")
  ADAO_Case.edAddChild(compute_bloc)
  ADAO_Case.edAddCFLink(CAS_node, compute_bloc)
  # We use an optimizer loop
  name = "Execute" + study_config["Algorithm"]
  if sys.version_info.major < 3:
    algLib = "daYacsIntegration.py"
  else:
    algLib = "adao.py"
  factoryName = "AssimilationAlgorithm_asynch"
  #~ optimizer_node = runtime.createOptimizerLoop(name, algLib, factoryName, "")
  optimizer_node = runtime.createOptimizerLoop(name, algLib, factoryName, True)
  compute_bloc.edAddChild(optimizer_node)
  ADAO_Case.edAddDFLink(CAS_node.getOutputPort("Study"), optimizer_node.edGetAlgoInitPort())

  # Check if we have a python script for OptimizerLoopNode
  data_config = study_config["ObservationOperator"]
  opt_script_nodeOO = None
  if data_config["Type"] == "Function" and (data_config["From"] == "ScriptWithSwitch" or data_config["From"] == "FunctionDict"):
    # Get script
    TheData = data_config["Data"]
    script_filename = ""
    for FunctionName in TheData["Function"]:
      # We currently support only one file
      script_filename = TheData["Script"][FunctionName]
      break
    # We create a new pyscript node
    opt_script_nodeOO = runtime.createScriptNode("", "FunctionNodeOO")
    if repertory and os.path.exists(os.path.join(base_repertory, os.path.basename(script_filename))):
      script_filename = os.path.join(base_repertory, os.path.basename(script_filename))
    try:
      script_str= open(script_filename, 'r')
    except:
      raise ValueError("Exception in opening function script file: " + script_filename)
    node_script  = "#-*- coding: utf-8 -*-\n"
    node_script += "import sys, os \n"
    node_script += "filepath = \"" + os.path.dirname(script_filename) + "\"\n"
    node_script += "filename = \"" + os.path.basename(script_filename) + "\"\n"
    node_script += "if sys.path.count(filepath)==0 or (sys.path.count(filepath)>0 and sys.path.index(filepath)>0):\n"
    node_script += "  sys.path.insert(0,filepath)\n"
    node_script += script_str.read()
    opt_script_nodeOO.setScript(node_script)
    opt_script_nodeOO.edAddInputPort("computation", t_param_input)
    opt_script_nodeOO.edAddOutputPort("result", t_param_output)

  elif data_config["Type"] == "Function" and data_config["From"] == "ScriptWithFunctions":
    # Get script
    ScriptWithFunctions = data_config["Data"]
    script_filename = ""
    for FunctionName in ScriptWithFunctions["Function"]:
      # We currently support only one file
      script_filename = ScriptWithFunctions["Script"][FunctionName]
      break

    # We create a new pyscript node
    opt_script_nodeOO = runtime.createScriptNode("", "FunctionNodeOO")
    if repertory and os.path.exists(os.path.join(base_repertory, os.path.basename(script_filename))):
      script_filename = os.path.join(base_repertory, os.path.basename(script_filename))
    try:
      script_str= open(script_filename, 'r')
    except:
      raise ValueError("Exception in opening function script file: " + script_filename)
    node_script  = "#-*- coding: utf-8 -*-\n"
    node_script += "import sys, os, numpy, logging, pickle, codecs\n"
    node_script += "def loads( data ):\n"
    node_script += "  return pickle.loads(codecs.decode(data.encode(), 'base64'))\n"
    node_script += "filepath = \"" + os.path.dirname(script_filename) + "\"\n"
    node_script += "filename = \"" + os.path.basename(script_filename) + "\"\n"
    node_script += "if sys.path.count(filepath)==0 or (sys.path.count(filepath)>0 and sys.path.index(filepath)>0):\n"
    node_script += "  sys.path.insert(0,filepath)\n"
    node_script += """# ==============================================\n"""
    node_script += script_str.read()
    node_script += """# ==============================================\n"""
    node_script += """__method = None\n"""
    node_script += """for param in computation["specificParameters"]:\n"""
    node_script += """  if param["name"] == "method": __method = loads(param["value"])\n"""
    node_script += """if __method not in ["Direct", "Tangent", "Adjoint"]:\n"""
    node_script += """  raise ValueError("ComputationFunctionNode: no valid computation method is given, it has to be Direct, Tangent or Adjoint (\'%s\' given)."%__method)\n"""
    node_script += """logging.debug("ComputationFunctionNode: Found method is \'%s\'"%__method)\n"""
    node_script += """#\n"""
    node_script += """#\n"""
    node_script += """__data = []\n"""
    node_script += """if __method == "Direct":\n"""
    node_script += """  try:\n"""
    node_script += """      DirectOperator\n"""
    node_script += """  except NameError:\n"""
    node_script += """      raise ValueError("ComputationFunctionNode: DirectOperator not found in the imported user script file")\n"""
    node_script += """  logging.debug("ComputationFunctionNode: Direct computation")\n"""
    node_script += """  __Xcurrent = computation["inputValues"][0][0][0]\n"""
    node_script += """  __data = DirectOperator(numpy.matrix( __Xcurrent ).T)\n"""
    node_script += """#\n"""
    node_script += """if __method == "Tangent":\n"""
    node_script += """  try:\n"""
    node_script += """    TangentOperator\n"""
    node_script += """  except NameError:\n"""
    node_script += """    raise ValueError("ComputationFunctionNode:  TangentOperator not found in the imported user script file")\n"""
    node_script += """  logging.debug("ComputationFunctionNode: Tangent computation")\n"""
    node_script += """  __Xcurrent  = computation["inputValues"][0][0][0]\n"""
    node_script += """  __dXcurrent = computation["inputValues"][0][0][1]\n"""
    node_script += """  __data = TangentOperator((numpy.matrix( __Xcurrent ).T, numpy.matrix( __dXcurrent ).T))\n"""
    node_script += """#\n"""
    node_script += """if __method == "Adjoint":\n"""
    node_script += """  try:\n"""
    node_script += """    AdjointOperator\n"""
    node_script += """  except NameError:\n"""
    node_script += """    raise ValueError("ComputationFunctionNode: AdjointOperator not found in the imported user script file")\n"""
    node_script += """  logging.debug("ComputationFunctionNode: Adjoint computation")\n"""
    node_script += """  __Xcurrent = computation["inputValues"][0][0][0]\n"""
    node_script += """  __Ycurrent = computation["inputValues"][0][0][1]\n"""
    node_script += """  __data = AdjointOperator((numpy.matrix( __Xcurrent ).T, numpy.matrix( __Ycurrent ).T))\n"""
    node_script += """#\n"""
    node_script += """logging.debug("ComputationFunctionNode: Formatting the output")\n"""
    node_script += """__it = 1.*numpy.ravel(__data)\n"""
    node_script += """outputValues = [[[[]]]]\n"""
    node_script += """outputValues[0][0][0] = list(__it)\n"""
    node_script += """#\n"""
    node_script += """result = {}\n"""
    node_script += """result["outputValues"]        = outputValues\n"""
    node_script += """result["specificOutputInfos"] = []\n"""
    node_script += """result["returnCode"]          = 0\n"""
    node_script += """result["errorMessage"]        = ""\n"""
    node_script += """# ==============================================\n"""
    #
    opt_script_nodeOO.setScript(node_script)
    opt_script_nodeOO.edAddInputPort("computation", t_param_input)
    opt_script_nodeOO.edAddOutputPort("result", t_param_output)

  elif data_config["Type"] == "Function" and data_config["From"] == "ScriptWithOneFunction":
    # Get script
    ScriptWithOneFunction = data_config["Data"]
    script_filename = ""
    for FunctionName in ScriptWithOneFunction["Function"]:
      # We currently support only one file
      script_filename = ScriptWithOneFunction["Script"][FunctionName]
      break

    # We create a new pyscript node
    opt_script_nodeOO = runtime.createScriptNode("", "FunctionNodeOO")
    if repertory and os.path.exists(os.path.join(base_repertory, os.path.basename(script_filename))):
      script_filename = os.path.join(base_repertory, os.path.basename(script_filename))
    try:
      script_str= open(script_filename, 'r')
    except:
      raise ValueError("Exception in opening function script file: " + script_filename)
    node_script  = "#-*- coding: utf-8 -*-\n"
    node_script += "import sys, os, numpy, logging, pickle, codecs\n"
    node_script += "def loads( data ):\n"
    node_script += "  return pickle.loads(codecs.decode(data.encode(), 'base64'))\n"
    node_script += "filepath = \"" + os.path.dirname(script_filename) + "\"\n"
    node_script += "filename = \"" + os.path.basename(script_filename) + "\"\n"
    node_script += "if sys.path.count(filepath)==0 or (sys.path.count(filepath)>0 and sys.path.index(filepath)>0):\n"
    node_script += "  sys.path.insert(0,filepath)\n"
    node_script += """# ==============================================\n"""
    node_script += script_str.read()
    node_script += """# ==============================================\n"""
    node_script += """__method = None\n"""
    node_script += """for param in computation["specificParameters"]:\n"""
    node_script += """  if param["name"] == "method": __method = loads(param["value"])\n"""
    node_script += """if __method not in ["Direct", "Tangent", "Adjoint"]:\n"""
    node_script += """  raise ValueError("ComputationFunctionNode: no valid computation method is given, it has to be Direct, Tangent or Adjoint (\'%s\' given)."%__method)\n"""
    node_script += """logging.debug("ComputationFunctionNode: Found method is \'%s\'"%__method)\n"""
    node_script += """#\n"""
    node_script += """try:\n"""
    node_script += """    DirectOperator\n"""
    node_script += """except NameError:\n"""
    node_script += """    raise ValueError("ComputationFunctionNode: DirectOperator not found in the imported user script file")\n"""
    if sys.version_info.major < 3:
        node_script += """from daCore import NumericObjects\n"""
    else:
        node_script += """from adao.daCore import NumericObjects\n"""
    node_script += """FDA = NumericObjects.FDApproximation(\n"""
    node_script += """    Function   = DirectOperator,\n"""
    node_script += """    increment  = %s,\n"""%str(ScriptWithOneFunction['DifferentialIncrement'])
    node_script += """    centeredDF = %s,\n"""%str(ScriptWithOneFunction['CenteredFiniteDifference'])
    if 'EnableMultiProcessing' in list(ScriptWithOneFunction.keys()):
        node_script += """    mpEnabled  = %s,\n"""%str(ScriptWithOneFunction['EnableMultiProcessing'])
    if 'NumberOfProcesses' in list(ScriptWithOneFunction.keys()):
        node_script += """    mpWorkers  = %s,\n"""%str(ScriptWithOneFunction['NumberOfProcesses'])
    node_script += """    )\n"""
    node_script += """#\n"""
    node_script += """__data = []\n"""
    node_script += """if __method == "Direct":\n"""
    node_script += """  logging.debug("ComputationFunctionNode: Direct computation")\n"""
    node_script += """  __Xcurrent = computation["inputValues"][0][0][0]\n"""
    node_script += """  __data = FDA.DirectOperator(numpy.matrix( __Xcurrent ).T)\n"""
    node_script += """#\n"""
    node_script += """if __method == "Tangent":\n"""
    node_script += """  logging.debug("ComputationFunctionNode: Tangent computation")\n"""
    node_script += """  __Xcurrent  = computation["inputValues"][0][0][0]\n"""
    node_script += """  __dXcurrent = computation["inputValues"][0][0][1]\n"""
    node_script += """  __data = FDA.TangentOperator((numpy.matrix( __Xcurrent ).T, numpy.matrix( __dXcurrent ).T))\n"""
    node_script += """#\n"""
    node_script += """if __method == "Adjoint":\n"""
    node_script += """  logging.debug("ComputationFunctionNode: Adjoint computation")\n"""
    node_script += """  __Xcurrent = computation["inputValues"][0][0][0]\n"""
    node_script += """  __Ycurrent = computation["inputValues"][0][0][1]\n"""
    node_script += """  __data = FDA.AdjointOperator((numpy.matrix( __Xcurrent ).T, numpy.matrix( __Ycurrent ).T))\n"""
    node_script += """#\n"""
    node_script += """logging.debug("ComputationFunctionNode: Formatting the output")\n"""
    node_script += """__it = 1.*numpy.ravel(__data)\n"""
    node_script += """outputValues = [[[[]]]]\n"""
    node_script += """outputValues[0][0][0] = list(__it)\n"""
    node_script += """#\n"""
    node_script += """result = {}\n"""
    node_script += """result["outputValues"]        = outputValues\n"""
    node_script += """result["specificOutputInfos"] = []\n"""
    node_script += """result["returnCode"]          = 0\n"""
    node_script += """result["errorMessage"]        = ""\n"""
    node_script += """# ==============================================\n"""
    #
    opt_script_nodeOO.setScript(node_script)
    opt_script_nodeOO.edAddInputPort("computation", t_param_input)
    opt_script_nodeOO.edAddOutputPort("result", t_param_output)

  else:
    factory_opt_script_node = catalogAd.getNodeFromNodeMap("FakeOptimizerLoopNode")
    opt_script_nodeOO = factory_opt_script_node.cloneNode("FakeFunctionNode")
  #
  if ExecuteInContainer and bool(UseSeparateContainer):
    opt_script_nodeOO.setExecutionMode(SALOMERuntime.PythonNode.REMOTE_STR)
    opt_script_nodeOO.setContainer(mycontainer)

  # Check if we have a python script for OptimizerLoopNode
  if "EvolutionModel" in list(study_config.keys()):
    data_config = study_config["EvolutionModel"]
    opt_script_nodeEM = None
    if data_config["Type"] == "Function" and (data_config["From"] == "ScriptWithSwitch" or data_config["From"] == "FunctionDict"):
      # Get script
      TheData = data_config["Data"]
      script_filename = ""
      for FunctionName in TheData["Function"]:
        # We currently support only one file
        script_filename = TheData["Script"][FunctionName]
        break
      # We create a new pyscript node
      opt_script_nodeEM = runtime.createScriptNode("", "FunctionNodeEM")
      if repertory and os.path.exists(os.path.join(base_repertory, os.path.basename(script_filename))):
        script_filename = os.path.join(base_repertory, os.path.basename(script_filename))
      try:
        script_str= open(script_filename, 'r')
      except:
        raise ValueError("Exception in opening function script file: " + script_filename)
      node_script  = "#-*- coding: utf-8 -*-\n"
      node_script += "import sys, os \n"
      node_script += "filepath = \"" + os.path.dirname(script_filename) + "\"\n"
      node_script += "filename = \"" + os.path.basename(script_filename) + "\"\n"
      node_script += "if sys.path.count(filepath)==0 or (sys.path.count(filepath)>0 and sys.path.index(filepath)>0):\n"
      node_script += "  sys.path.insert(0,filepath)\n"
      node_script += script_str.read()
      opt_script_nodeEM.setScript(node_script)
      opt_script_nodeEM.edAddInputPort("computation", t_param_input)
      opt_script_nodeEM.edAddOutputPort("result", t_param_output)

    elif data_config["Type"] == "Function" and data_config["From"] == "ScriptWithFunctions":
      # Get script
      ScriptWithFunctions = data_config["Data"]
      script_filename = ""
      for FunctionName in ScriptWithFunctions["Function"]:
        # We currently support only one file
        script_filename = ScriptWithFunctions["Script"][FunctionName]
        break
      # We create a new pyscript node
      opt_script_nodeEM = runtime.createScriptNode("", "FunctionNodeEM")
      if repertory and os.path.exists(os.path.join(base_repertory, os.path.basename(script_filename))):
        script_filename = os.path.join(base_repertory, os.path.basename(script_filename))
      try:
        script_str= open(script_filename, 'r')
      except:
        raise ValueError("Exception in opening function script file: " + script_filename)
      node_script  = "#-*- coding: utf-8 -*-\n"
      node_script += "import sys, os, numpy, logging, pickle, codecs\n"
      node_script += "def loads( data ):\n"
      node_script += "  return pickle.loads(codecs.decode(data.encode(), 'base64'))\n"
      node_script += "filepath = \"" + os.path.dirname(script_filename) + "\"\n"
      node_script += "filename = \"" + os.path.basename(script_filename) + "\"\n"
      node_script += "if sys.path.count(filepath)==0 or (sys.path.count(filepath)>0 and sys.path.index(filepath)>0):\n"
      node_script += "  sys.path.insert(0,filepath)\n"
      node_script += script_str.read()
      node_script += """# ==============================================\n"""
      node_script += """__method = None\n"""
      node_script += """for param in computation["specificParameters"]:\n"""
      node_script += """  if param["name"] == "method": __method = loads(param["value"])\n"""
      node_script += """if __method not in ["Direct", "Tangent", "Adjoint"]:\n"""
      node_script += """  raise ValueError("ComputationFunctionNode: no valid computation method is given, it has to be Direct, Tangent or Adjoint (\'%s\' given)."%__method)\n"""
      node_script += """logging.debug("ComputationFunctionNode: Found method is \'%s\'"%__method)\n"""
      node_script += """#\n"""
      node_script += """#\n"""
      node_script += """__data = []\n"""
      node_script += """if __method == "Direct":\n"""
      node_script += """  try:\n"""
      node_script += """    DirectOperator\n"""
      node_script += """  except NameError:\n"""
      node_script += """    raise ValueError("ComputationFunctionNode: mandatory DirectOperator not found in the imported user script file")\n"""
      node_script += """  logging.debug("ComputationFunctionNode: Direct computation")\n"""
      node_script += """  __Xcurrent = computation["inputValues"][0][0][0]\n"""
      node_script += """  if len(computation["inputValues"][0][0]) == 2:\n"""
      node_script += """    __Ucurrent = computation["inputValues"][0][0][1]\n"""
      node_script += """    __data = DirectOperator((numpy.matrix( __Xcurrent ).T, numpy.matrix( __Ucurrent ).T))\n"""
      node_script += """  else:\n"""
      node_script += """    __data = DirectOperator(numpy.matrix( __Xcurrent ).T)\n"""
      node_script += """#\n"""
      node_script += """if __method == "Tangent":\n"""
      node_script += """  try:\n"""
      node_script += """    TangentOperator\n"""
      node_script += """  except NameError:\n"""
      node_script += """    raise ValueError("ComputationFunctionNode: mandatory TangentOperator not found in the imported user script file")\n"""
      node_script += """  logging.debug("ComputationFunctionNode: Tangent computation")\n"""
      node_script += """  __Xcurrent  = computation["inputValues"][0][0][0]\n"""
      node_script += """  __dXcurrent = computation["inputValues"][0][0][1]\n"""
      node_script += """  __data = TangentOperator((numpy.matrix( __Xcurrent ).T, numpy.matrix( __dXcurrent ).T))\n"""
      node_script += """#\n"""
      node_script += """if __method == "Adjoint":\n"""
      node_script += """  try:\n"""
      node_script += """    AdjointOperator\n"""
      node_script += """  except NameError:\n"""
      node_script += """    raise ValueError("ComputationFunctionNode: mandatory AdjointOperator not found in the imported user script file")\n"""
      node_script += """  logging.debug("ComputationFunctionNode: Adjoint computation")\n"""
      node_script += """  __Xcurrent = computation["inputValues"][0][0][0]\n"""
      node_script += """  __Ycurrent = computation["inputValues"][0][0][1]\n"""
      node_script += """  __data = AdjointOperator((numpy.matrix( __Xcurrent ).T, numpy.matrix( __Ycurrent ).T))\n"""
      node_script += """#\n"""
      node_script += """logging.debug("ComputationFunctionNode: Formatting the output")\n"""
      node_script += """__it = 1.*numpy.ravel(__data)\n"""
      node_script += """outputValues = [[[[]]]]\n"""
      node_script += """outputValues[0][0][0] = list(__it)\n"""
      node_script += """#\n"""
      node_script += """result = {}\n"""
      node_script += """result["outputValues"]        = outputValues\n"""
      node_script += """result["specificOutputInfos"] = []\n"""
      node_script += """result["returnCode"]          = 0\n"""
      node_script += """result["errorMessage"]        = ""\n"""
      node_script += """# ==============================================\n"""
      #
      opt_script_nodeEM.setScript(node_script)
      opt_script_nodeEM.edAddInputPort("computation", t_param_input)
      opt_script_nodeEM.edAddOutputPort("result", t_param_output)

    elif data_config["Type"] == "Function" and data_config["From"] == "ScriptWithOneFunction":
      # Get script
      ScriptWithOneFunction = data_config["Data"]
      script_filename = ""
      for FunctionName in ScriptWithOneFunction["Function"]:
        # We currently support only one file
        script_filename = ScriptWithOneFunction["Script"][FunctionName]
        break
      # We create a new pyscript node
      opt_script_nodeEM = runtime.createScriptNode("", "FunctionNodeEM")
      if repertory and os.path.exists(os.path.join(base_repertory, os.path.basename(script_filename))):
        script_filename = os.path.join(base_repertory, os.path.basename(script_filename))
      try:
        script_str= open(script_filename, 'r')
      except:
        raise ValueError("Exception in opening function script file: " + script_filename)
      node_script  = "#-*- coding: utf-8 -*-\n"
      node_script += "import sys, os, numpy, logging, pickle, codecs\n"
      node_script += "def loads( data ):\n"
      node_script += "  return pickle.loads(codecs.decode(data.encode(), 'base64'))\n"
      node_script += "filepath = \"" + os.path.dirname(script_filename) + "\"\n"
      node_script += "filename = \"" + os.path.basename(script_filename) + "\"\n"
      node_script += "if sys.path.count(filepath)==0 or (sys.path.count(filepath)>0 and sys.path.index(filepath)>0):\n"
      node_script += "  sys.path.insert(0,filepath)\n"
      node_script += script_str.read()
      node_script += """# ==============================================\n"""
      node_script += """__method = None\n"""
      node_script += """for param in computation["specificParameters"]:\n"""
      node_script += """  if param["name"] == "method": __method = loads(param["value"])\n"""
      node_script += """if __method not in ["Direct", "Tangent", "Adjoint"]:\n"""
      node_script += """  raise ValueError("ComputationFunctionNode: no valid computation method is given, it has to be Direct, Tangent or Adjoint (\'%s\' given)."%__method)\n"""
      node_script += """logging.debug("ComputationFunctionNode: Found method is \'%s\'"%__method)\n"""
      node_script += """#\n"""
      node_script += """try:\n"""
      node_script += """    DirectOperator\n"""
      node_script += """except NameError:\n"""
      node_script += """    raise ValueError("ComputationFunctionNode: DirectOperator not found in the imported user script file")\n"""
      if sys.version_info.major < 3:
          node_script += """from daCore import NumericObjects\n"""
      else:
          node_script += """from adao.daCore import NumericObjects\n"""
      node_script += """FDA = NumericObjects.FDApproximation(\n"""
      node_script += """    Function   = DirectOperator,\n"""
      node_script += """    increment  = %s,\n"""%str(ScriptWithOneFunction['DifferentialIncrement'])
      node_script += """    centeredDF = %s,\n"""%str(ScriptWithOneFunction['CenteredFiniteDifference'])
      if 'EnableMultiProcessing' in list(ScriptWithOneFunction.keys()):
          node_script += """    mpEnabled  = %s,\n"""%str(ScriptWithOneFunction['EnableMultiProcessing'])
      if 'NumberOfProcesses' in list(ScriptWithOneFunction.keys()):
          node_script += """    mpWorkers  = %s,\n"""%str(ScriptWithOneFunction['NumberOfProcesses'])
      node_script += """    )\n"""
      node_script += """#\n"""
      node_script += """__data = []\n"""
      node_script += """if __method == "Direct":\n"""
      node_script += """  logging.debug("ComputationFunctionNode: Direct computation")\n"""
      node_script += """  if len(computation["inputValues"][0][0]) == 2:\n"""
      node_script += """    raise ValueError("ComputationFunctionNode: you have to build explicitly the controled evolution model and its tangent and adjoint operators, instead of using approximate derivative.")"""
      node_script += """  __Xcurrent = computation["inputValues"][0][0][0]\n"""
      node_script += """  __data = FDA.DirectOperator(numpy.matrix( __Xcurrent ).T)\n"""
      node_script += """#\n"""
      node_script += """if __method == "Tangent":\n"""
      node_script += """  logging.debug("ComputationFunctionNode: Tangent computation")\n"""
      node_script += """  __Xcurrent  = computation["inputValues"][0][0][0]\n"""
      node_script += """  __dXcurrent = computation["inputValues"][0][0][1]\n"""
      node_script += """  __data = FDA.TangentOperator((numpy.matrix( __Xcurrent ).T, numpy.matrix( __dXcurrent ).T))\n"""
      node_script += """#\n"""
      node_script += """if __method == "Adjoint":\n"""
      node_script += """  logging.debug("ComputationFunctionNode: Adjoint computation")\n"""
      node_script += """  __Xcurrent = computation["inputValues"][0][0][0]\n"""
      node_script += """  __Ycurrent = computation["inputValues"][0][0][1]\n"""
      node_script += """  __data = FDA.AdjointOperator((numpy.matrix( __Xcurrent ).T, numpy.matrix( __Ycurrent ).T))\n"""
      node_script += """#\n"""
      node_script += """logging.debug("ComputationFunctionNode: Formatting the output")\n"""
      node_script += """__it = 1.*numpy.ravel(__data)\n"""
      node_script += """outputValues = [[[[]]]]\n"""
      node_script += """outputValues[0][0][0] = list(__it)\n"""
      node_script += """#\n"""
      node_script += """result = {}\n"""
      node_script += """result["outputValues"]        = outputValues\n"""
      node_script += """result["specificOutputInfos"] = []\n"""
      node_script += """result["returnCode"]          = 0\n"""
      node_script += """result["errorMessage"]        = ""\n"""
      node_script += """# ==============================================\n"""
      #
      opt_script_nodeEM.setScript(node_script)
      opt_script_nodeEM.edAddInputPort("computation", t_param_input)
      opt_script_nodeEM.edAddOutputPort("result", t_param_output)

    else:
      factory_opt_script_node = catalogAd.getNodeFromNodeMap("FakeOptimizerLoopNode")
      opt_script_nodeEM = factory_opt_script_node.cloneNode("FakeFunctionNode")
    #
    if ExecuteInContainer and bool(UseSeparateContainer):
      opt_script_nodeEM.setExecutionMode(SALOMERuntime.PythonNode.REMOTE_STR)
      opt_script_nodeEM.setContainer(mycontainer)

  # Add computation bloc
  if "Observers" in list(study_config.keys()):
    execution_bloc = runtime.createBloc("Execution Bloc")
    optimizer_node.edSetNode(execution_bloc)

    # Add a node that permits to configure the switch
    factory_read_for_switch_node = catalogAd.getNodeFromNodeMap("ReadForSwitchNode")
    read_for_switch_node = factory_read_for_switch_node.cloneNode("ReadForSwitch")
    if ExecuteInContainer and bool(UseSeparateContainer):
      read_for_switch_node.setExecutionMode(SALOMERuntime.PythonNode.REMOTE_STR)
      read_for_switch_node.setContainer(mycontainer)
    execution_bloc.edAddChild(read_for_switch_node)
    ADAO_Case.edAddDFLink(optimizer_node.edGetSamplePort(), read_for_switch_node.getInputPort("data"))

    # Add a switch
    switch_node = runtime.createSwitch("Execution Switch")
    execution_bloc.edAddChild(switch_node)
    # Connect switch
    ADAO_Case.edAddDFLink(read_for_switch_node.getOutputPort("switch_value"), switch_node.edGetConditionPort())

    # First case: computation bloc
    computation_blocOO = runtime.createBloc("computation_blocOO")
    computation_blocOO.edAddChild(opt_script_nodeOO)
    switch_node.edSetNode(1, computation_blocOO)

    # We connect with the script
    ADAO_Case.edAddDFLink(read_for_switch_node.getOutputPort("data"), opt_script_nodeOO.getInputPort("computation"))
    ADAO_Case.edAddDFLink(opt_script_nodeOO.getOutputPort("result"), optimizer_node.edGetPortForOutPool())

    # Second case: evolution bloc
    if "EvolutionModel" in list(study_config.keys()):
      computation_blocEM = runtime.createBloc("computation_blocEM")
      computation_blocEM.edAddChild(opt_script_nodeEM)
      switch_node.edSetNode(2, computation_blocEM)

      # We connect with the script
      ADAO_Case.edAddDFLink(read_for_switch_node.getOutputPort("data"), opt_script_nodeEM.getInputPort("computation"))
      ADAO_Case.edAddDFLink(opt_script_nodeEM.getOutputPort("result"), optimizer_node.edGetPortForOutPool())

    # For each observer add a new bloc in the switch
    observer_config = study_config["Observers"]
    for observer_name in observer_config:
      observer_cfg = observer_config[observer_name]
      observer_bloc = runtime.createBloc("Observer %s" % observer_name)
      switch_node.edSetNode(observer_cfg["number"], observer_bloc)

      factory_extract_data_node = catalogAd.getNodeFromNodeMap("ExtractDataNode")
      extract_data_node = factory_extract_data_node.cloneNode("ExtractData")
      if ExecuteInContainer and bool(UseSeparateContainer):
        extract_data_node.setExecutionMode(SALOMERuntime.PythonNode.REMOTE_STR)
        extract_data_node.setContainer(mycontainer)
      observer_bloc.edAddChild(extract_data_node)
      ADAO_Case.edAddDFLink(read_for_switch_node.getOutputPort("data"), extract_data_node.getInputPort("data"))

      observation_node = None
      if observer_cfg["nodetype"] == "String":
        factory_observation_node = catalogAd.getNodeFromNodeMap("ObservationNodeString")
        observation_node = factory_observation_node.cloneNode("Observation")
        node_script = observation_node.getScript()
        node_script += observer_cfg["String"]
        observation_node.setScript(node_script)
      else:
        factory_observation_node = catalogAd.getNodeFromNodeMap("ObservationNodeFile")
        observation_node = factory_observation_node.cloneNode("Observation")
        _Internal_Add_dir_script_ports( observation_node, observer_cfg["Script"], repertory, base_repertory, t_string)
      if ExecuteInContainer and bool(UseSeparateContainer):
        observation_node.setExecutionMode(SALOMERuntime.PythonNode.REMOTE_STR)
        observation_node.setContainer(mycontainer)
      observer_bloc.edAddChild(observation_node)
      ADAO_Case.edAddDFLink(extract_data_node.getOutputPort("var"), observation_node.getInputPort("var"))
      ADAO_Case.edAddDFLink(extract_data_node.getOutputPort("info"), observation_node.getInputPort("info"))

      factory_end_observation_node = catalogAd.getNodeFromNodeMap("EndObservationNode")
      end_observation_node = factory_end_observation_node.cloneNode("EndObservation")
      if ExecuteInContainer and bool(UseSeparateContainer):
        end_observation_node.setExecutionMode(SALOMERuntime.PythonNode.REMOTE_STR)
        end_observation_node.setContainer(mycontainer)
      observer_bloc.edAddChild(end_observation_node)
      ADAO_Case.edAddCFLink(observation_node, end_observation_node)
      ADAO_Case.edAddDFLink(end_observation_node.getOutputPort("output"), optimizer_node.edGetPortForOutPool())

  elif "EvolutionModel" in list(study_config.keys()):
    execution_bloc = runtime.createBloc("Execution Bloc")
    optimizer_node.edSetNode(execution_bloc)

    # Add a node that permits to configure the switch
    factory_read_for_switch_node = catalogAd.getNodeFromNodeMap("ReadForSwitchNode")
    read_for_switch_node = factory_read_for_switch_node.cloneNode("ReadForSwitch")
    if ExecuteInContainer and bool(UseSeparateContainer):
      read_for_switch_node.setExecutionMode(SALOMERuntime.PythonNode.REMOTE_STR)
      read_for_switch_node.setContainer(mycontainer)
    execution_bloc.edAddChild(read_for_switch_node)
    ADAO_Case.edAddDFLink(optimizer_node.edGetSamplePort(), read_for_switch_node.getInputPort("data"))

    # Add a switch
    switch_node = runtime.createSwitch("Execution Switch")
    execution_bloc.edAddChild(switch_node)
    # Connect switch
    ADAO_Case.edAddDFLink(read_for_switch_node.getOutputPort("switch_value"), switch_node.edGetConditionPort())

    # First case: computation bloc
    computation_blocOO = runtime.createBloc("computation_blocOO")
    computation_blocOO.edAddChild(opt_script_nodeOO)
    switch_node.edSetNode(1, computation_blocOO)

    # We connect with the script
    ADAO_Case.edAddDFLink(read_for_switch_node.getOutputPort("data"), opt_script_nodeOO.getInputPort("computation"))
    ADAO_Case.edAddDFLink(opt_script_nodeOO.getOutputPort("result"), optimizer_node.edGetPortForOutPool())

    # Second case: evolution bloc
    computation_blocEM = runtime.createBloc("computation_blocEM")
    computation_blocEM.edAddChild(opt_script_nodeEM)
    switch_node.edSetNode(2, computation_blocEM)

    # We connect with the script
    ADAO_Case.edAddDFLink(read_for_switch_node.getOutputPort("data"), opt_script_nodeEM.getInputPort("computation"))
    ADAO_Case.edAddDFLink(opt_script_nodeEM.getOutputPort("result"), optimizer_node.edGetPortForOutPool())

  else:
    computation_blocOO = runtime.createBloc("computation_blocOO")
    optimizer_node.edSetNode(computation_blocOO)
    computation_blocOO.edAddChild(opt_script_nodeOO)

    # We connect Optimizer with the script
    ADAO_Case.edAddDFLink(optimizer_node.edGetSamplePort(), opt_script_nodeOO.getInputPort("computation"))
    ADAO_Case.edAddDFLink(opt_script_nodeOO.getOutputPort("result"), optimizer_node.edGetPortForOutPool())

  # Connect node with InitUserData
  if "ObservationOperator" in init_config["Target"]:
    opt_node_script = opt_script_nodeOO.getScript()
    opt_node_script = "__builtins__[\"init_data\"] = init_data\n" + opt_node_script
    opt_script_nodeOO.setScript(opt_node_script)
    opt_script_nodeOO.edAddInputPort("init_data", t_pyobj)
    ADAO_Case.edAddDFLink(init_node.getOutputPort("init_data"), opt_script_nodeOO.getInputPort("init_data"))

  # Step 4: create post-processing from user configuration
  if "UserPostAnalysis" in list(study_config.keys()):
    analysis_config = study_config["UserPostAnalysis"]
    if analysis_config["From"] == "String":
      factory_analysis_node = catalogAd.getNodeFromNodeMap("SimpleUserAnalysis")
      analysis_node = factory_analysis_node.cloneNode("UsePostAnalysis")
      if ExecuteInContainer and bool(UseSeparateContainer):
        analysis_node.setExecutionMode(SALOMERuntime.PythonNode.REMOTE_STR)
        analysis_node.setContainer(mycontainer)
      default_script = analysis_node.getScript()
      final_script = default_script + analysis_config["Data"]
      analysis_node.setScript(final_script)
      ADAO_Case.edAddChild(analysis_node)
      ADAO_Case.edAddCFLink(compute_bloc, analysis_node)
      if AlgoType[study_config["Algorithm"]] == "Optim":
        ADAO_Case.edAddDFLink(optimizer_node.edGetAlgoResultPort(), analysis_node.getInputPort("Study"))
      else:
        ADAO_Case.edAddDFLink(execute_node.getOutputPort("Study"), analysis_node.getInputPort("Study"))

      # Connect node with InitUserData
      if "UserPostAnalysis" in init_config["Target"]:
        node_script = analysis_node.getScript()
        node_script = "__builtins__[\"init_data\"] = init_data\n" + node_script
        analysis_node.setScript(node_script)
        analysis_node.edAddInputPort("init_data", t_pyobj)
        ADAO_Case.edAddDFLink(init_node.getOutputPort("init_data"), analysis_node.getInputPort("init_data"))

    elif analysis_config["From"] == "Script":
      factory_analysis_node = catalogAd.getNodeFromNodeMap("SimpleUserAnalysis")
      analysis_node = factory_analysis_node.cloneNode("UserPostAnalysis")
      if ExecuteInContainer and bool(UseSeparateContainer):
        analysis_node.setExecutionMode(SALOMERuntime.PythonNode.REMOTE_STR)
        analysis_node.setContainer(mycontainer)
      default_script = analysis_node.getScript()
      analysis_file_name = analysis_config["Data"]
      if repertory and os.path.exists(os.path.join(base_repertory, os.path.basename(analysis_file_name))):
        analysis_file_name = os.path.join(base_repertory, os.path.basename(analysis_file_name))
      try:
        analysis_file = open(analysis_file_name, 'r')
      except:
        raise ValueError("Exception in opening analysis file: " + str(analysis_config["Data"]))
      node_script  = "#-*- coding: utf-8 -*-\n"
      node_script += "import sys, os \n"
      node_script += "filepath = \"" + os.path.dirname(analysis_file_name) + "\"\n"
      node_script += "filename = \"" + os.path.basename(analysis_file_name) + "\"\n"
      node_script += "if sys.path.count(filepath)==0 or (sys.path.count(filepath)>0 and sys.path.index(filepath)>0):\n"
      node_script += "  sys.path.insert(0,filepath)\n"
      node_script += default_script
      node_script += analysis_file.read()
      analysis_node.setScript(node_script)
      ADAO_Case.edAddChild(analysis_node)
      ADAO_Case.edAddCFLink(compute_bloc, analysis_node)
      if AlgoType[study_config["Algorithm"]] == "Optim":
        ADAO_Case.edAddDFLink(optimizer_node.edGetAlgoResultPort(), analysis_node.getInputPort("Study"))
      else:
        ADAO_Case.edAddDFLink(execute_node.getOutputPort("Study"), analysis_node.getInputPort("Study"))
      # Connect node with InitUserData
      if "UserPostAnalysis" in init_config["Target"]:
        node_script = analysis_node.getScript()
        node_script = "__builtins__[\"init_data\"] = init_data\n" + node_script
        analysis_node.setScript(node_script)
        analysis_node.edAddInputPort("init_data", t_pyobj)
        ADAO_Case.edAddDFLink(init_node.getOutputPort("init_data"), analysis_node.getInputPort("init_data"))

      pass

  return proc

def write_yacs_proc(proc, yacs_schema_filename):

  proc.saveSchema(yacs_schema_filename)




#-*- coding: utf-8 -*-
#
# Copyright (C) 2008-2022 EDF R&D
#
# This file is part of SALOME ADAO module
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
#
# See http://www.salome-platform.org/ or email : webmaster.salome@opencascade.com
#
# Author: Andre Ribes, andre.ribes@edf.fr, EDF R&D


# -- Infos pour le parser --

AnalysisData = {}
AnalysisFromList = ["String", "Script"]

# -- Infos from daCore --
AssimData = [
    "Background",
    "BackgroundError",
    "Observation",
    "ObservationError",
    "ObservationOperator",
    "EvolutionModel",
    "EvolutionError",
    "AlgorithmParameters",
    "CheckingPoint",
    "ControlInput",
    ]

AssimType = {}
AssimType["Background"]          = ["Vector", "VectorSerie"]
AssimType["BackgroundError"]     = ["Matrix", "ScalarSparseMatrix", "DiagonalSparseMatrix"]
AssimType["Observation"]         = ["Vector", "VectorSerie"]
AssimType["ObservationError"]    = ["Matrix", "ScalarSparseMatrix", "DiagonalSparseMatrix"]
AssimType["ObservationOperator"] = ["Matrix", "Function"]
AssimType["EvolutionModel"]      = ["Matrix", "Function"]
AssimType["EvolutionError"]      = ["Matrix", "ScalarSparseMatrix", "DiagonalSparseMatrix"]
AssimType["AlgorithmParameters"] = ["Dict"]
AssimType["UserDataInit"]        = ["Dict"]
AssimType["CheckingPoint"]       = ["Vector"]
AssimType["ControlInput"]        = ["Vector", "VectorSerie"]

FromNumpyList = {}
FromNumpyList["Vector"]               = ["String", "Script", "DataFile"]
FromNumpyList["VectorSerie"]          = ["String", "Script", "DataFile"]
FromNumpyList["Matrix"]               = ["String", "Script"]
FromNumpyList["ScalarSparseMatrix"]   = ["String", "Script"]
FromNumpyList["DiagonalSparseMatrix"] = ["String", "Script"]
FromNumpyList["Function"]             = ["ScriptWithOneFunction", "ScriptWithFunctions", "ScriptWithSwitch", "FunctionDict"]
FromNumpyList["Dict"]                 = ["String", "Script"]

# -- Infos from daAlgorithms --
AssimAlgos = [
    "3DVAR",
    "4DVAR",
    "Blue",
    "DerivativeFreeOptimization",
    "DifferentialEvolution",
    "EnsembleBlue",
    "EnsembleKalmanFilter",
    "ExtendedBlue",
    "ExtendedKalmanFilter",
    "KalmanFilter",
    "LinearLeastSquares",
    "NonLinearLeastSquares",
    "ParticleSwarmOptimization",
    "QuantileRegression",
    "TabuSearch",
    "UnscentedKalmanFilter",
    ]
OptimizationAlgos = [
    "Blue",
    "3DVAR",
    "DerivativeFreeOptimization",
    "DifferentialEvolution",
    "KalmanFilter",
    "LinearLeastSquares",
    "NonLinearLeastSquares",
    "ParticleSwarmOptimization",
    "TabuSearch",
    ]
ReductionAlgos = [
    "EnsembleBlue",
    "EnsembleKalmanFilter",
    "UnscentedKalmanFilter",
    ]
CheckAlgos = [
    "FunctionTest",
    "LinearityTest",
    "GradientTest",
    "AdjointTest",
    "TangentTest",
    "LocalSensitivityTest",
    "SamplingTest",
    "ParallelFunctionTest",
    "InputValuesTest",
    "ObserverTest",
    ]

AlgoDataRequirements = {}
AlgoDataRequirements["3DVAR"] = [
    "Background", "BackgroundError",
    "Observation", "ObservationError",
    "ObservationOperator",
    ]
AlgoDataRequirements["4DVAR"] = [
    "Background", "BackgroundError",
    "Observation", "ObservationError",
    "ObservationOperator",
    ]
AlgoDataRequirements["Blue"] = [
    "Background", "BackgroundError",
    "Observation", "ObservationError",
    "ObservationOperator",
    ]
AlgoDataRequirements["ExtendedBlue"] = [
    "Background", "BackgroundError",
    "Observation", "ObservationError",
    "ObservationOperator",
    ]
AlgoDataRequirements["EnsembleBlue"] = [
    "Background", "BackgroundError",
    "Observation", "ObservationError",
    "ObservationOperator",
    ]
AlgoDataRequirements["KalmanFilter"] = [
    "Background", "BackgroundError",
    "Observation", "ObservationError",
    "ObservationOperator",
    ]
AlgoDataRequirements["ExtendedKalmanFilter"] = [
    "Background", "BackgroundError",
    "Observation", "ObservationError",
    "ObservationOperator",
    ]
AlgoDataRequirements["EnsembleKalmanFilter"] = [
    "Background", "BackgroundError",
    "Observation", "ObservationError",
    "ObservationOperator",
    ]
AlgoDataRequirements["UnscentedKalmanFilter"] = [
    "Background", "BackgroundError",
    "Observation", "ObservationError",
    "ObservationOperator",
    ]
AlgoDataRequirements["QuantileRegression"] = [
    "Background",
    "Observation",
    "ObservationOperator",
    ]
AlgoDataRequirements["DerivativeFreeOptimization"] = [
    "Background", "BackgroundError",
    "Observation", "ObservationError",
    "ObservationOperator",
    ]
AlgoDataRequirements["ParticleSwarmOptimization"] = [
    "Background", "BackgroundError",
    "Observation", "ObservationError",
    "ObservationOperator",
    ]
AlgoDataRequirements["DifferentialEvolution"] = [
    "Background", "BackgroundError",
    "Observation", "ObservationError",
    "ObservationOperator",
    ]
AlgoDataRequirements["TabuSearch"] = [
    "Background", "BackgroundError",
    "Observation", "ObservationError",
    "ObservationOperator",
    ]
AlgoDataRequirements["LinearLeastSquares"] = [
    "Observation", "ObservationError",
    "ObservationOperator",
    ]
AlgoDataRequirements["NonLinearLeastSquares"] = [
    "Background",
    "Observation", "ObservationError",
    "ObservationOperator",
    ]

AlgoDataRequirements["FunctionTest"] = [
    "CheckingPoint",
    "ObservationOperator",
    ]
AlgoDataRequirements["LinearityTest"] = [
    "CheckingPoint",
    "ObservationOperator",
    ]
AlgoDataRequirements["GradientTest"] = [
    "CheckingPoint",
    "ObservationOperator",
    ]
AlgoDataRequirements["AdjointTest"] = [
    "CheckingPoint",
    "ObservationOperator",
    ]
AlgoDataRequirements["TangentTest"] = [
    "CheckingPoint",
    "ObservationOperator",
    ]
AlgoDataRequirements["LocalSensitivityTest"] = [
    "CheckingPoint",
    "Observation",
    "ObservationOperator",
    ]
AlgoDataRequirements["SamplingTest"] = [
    "CheckingPoint", "BackgroundError",
    "Observation", "ObservationError",
    "ObservationOperator",
    ]
AlgoDataRequirements["ParallelFunctionTest"] = [
    "CheckingPoint",
    "ObservationOperator",
    ]
AlgoDataRequirements["ObserverTest"] = [
    "Observers",
    ]
AlgoDataRequirements["InputValuesTest"] = [
    "CheckingPoint",
    ]

AlgoType = {}
AlgoType["3DVAR"] = "Optim"
AlgoType["4DVAR"] = "Optim"
AlgoType["Blue"] = "Optim"
AlgoType["ExtendedBlue"] = "Optim"
AlgoType["EnsembleBlue"] = "Optim"
AlgoType["KalmanFilter"] = "Optim"
AlgoType["ExtendedKalmanFilter"] = "Optim"
AlgoType["EnsembleKalmanFilter"] = "Optim"
AlgoType["UnscentedKalmanFilter"] = "Optim"
AlgoType["QuantileRegression"] = "Optim"
AlgoType["DerivativeFreeOptimization"] = "Optim"
AlgoType["ParticleSwarmOptimization"] = "Optim"
AlgoType["DifferentialEvolution"] = "Optim"
AlgoType["TabuSearch"] = "Optim"
AlgoType["LinearLeastSquares"] = "Optim"
AlgoType["NonLinearLeastSquares"] = "Optim"

# Variables qui sont partages avec le generateur de
# catalogue Eficas

# Basic data types
BasicDataInputs = ["String", "Script", "ScriptWithOneFunction", "ScriptWithFunctions", "ScriptWithSwitch", "FunctionDict"]

# Data input dict
DataTypeDict = {}
DataTypeDict["Vector"]               = ["String", "Script", "DataFile"]
DataTypeDict["VectorSerie"]          = ["String", "Script", "DataFile"]
DataTypeDict["Matrix"]               = ["String", "Script"]
DataTypeDict["ScalarSparseMatrix"]   = ["String", "Script"]
DataTypeDict["DiagonalSparseMatrix"] = ["String", "Script"]
DataTypeDict["Function"]             = ["ScriptWithOneFunction", "ScriptWithFunctions", "ScriptWithSwitch", "FunctionDict"]
DataTypeDict["Dict"]                 = ["String", "Script"]

DataTypeDefaultDict = {}
DataTypeDefaultDict["Vector"]               = "Script"
DataTypeDefaultDict["VectorSerie"]          = "Script"
DataTypeDefaultDict["Matrix"]               = "Script"
DataTypeDefaultDict["ScalarSparseMatrix"]   = "String"
DataTypeDefaultDict["DiagonalSparseMatrix"] = "String"
DataTypeDefaultDict["Function"]             = "ScriptWithOneFunction"
DataTypeDefaultDict["Dict"]                 = "Script"

DataSValueDefaultDict = {}
DataSValueDefaultDict["ScalarSparseMatrix"]   = "1."

# Assimilation data input
AssimDataDict = {}
AssimDataDict["Background"]          = ["Vector", "VectorSerie"]
AssimDataDict["BackgroundError"]     = ["Matrix", "ScalarSparseMatrix", "DiagonalSparseMatrix"]
AssimDataDict["Observation"]         = ["Vector", "VectorSerie"]
AssimDataDict["ObservationError"]    = ["Matrix", "ScalarSparseMatrix", "DiagonalSparseMatrix"]
AssimDataDict["ObservationOperator"] = ["Matrix", "Function"]
AssimDataDict["EvolutionModel"]      = ["Matrix", "Function"]
AssimDataDict["EvolutionError"]      = ["Matrix", "ScalarSparseMatrix", "DiagonalSparseMatrix"]
AssimDataDict["AlgorithmParameters"] = ["Dict"]
AssimDataDict["UserDataInit"]        = ["Dict"]
AssimDataDict["CheckingPoint"]       = ["Vector"]
AssimDataDict["ControlInput"]        = ["Vector", "VectorSerie"]

AssimDataDefaultDict = {}
AssimDataDefaultDict["Background"]          = "Vector"
AssimDataDefaultDict["BackgroundError"]     = "ScalarSparseMatrix"
AssimDataDefaultDict["Observation"]         = "Vector"
AssimDataDefaultDict["ObservationError"]    = "ScalarSparseMatrix"
AssimDataDefaultDict["ObservationOperator"] = "Function"
AssimDataDefaultDict["EvolutionModel"]      = "Function"
AssimDataDefaultDict["EvolutionError"]      = "ScalarSparseMatrix"
AssimDataDefaultDict["AlgorithmParameters"] = "Dict"
AssimDataDefaultDict["UserDataInit"]        = "Dict"
AssimDataDefaultDict["CheckingPoint"]       = "Vector"
AssimDataDefaultDict["ControlInput"]        = "Vector"

StoredAssimData = ["Vector", "VectorSerie", "Matrix", "ScalarSparseMatrix", "DiagonalSparseMatrix"]

# Assimilation optional nodes
OptDict = {}
OptDict["UserPostAnalysis"]   = ["String", "Script", "Template"]
OptDefaultDict = {}
OptDefaultDict["UserPostAnalysis"]   = "Template"

# Observers
ObserversList = [
    "Analysis",
    "Innovation",
    "InnovationAtCurrentState",
    "CurrentState",
    "CurrentOptimum",
    "IndexOfOptimum",
    "SimulatedObservationAtBackground",
    "SimulatedObservationAtCurrentState",
    "SimulatedObservationAtOptimum",
    "SimulatedObservationAtCurrentOptimum",
    "BMA",
    "OMA",
    "OMB",
    "CostFunctionJ",
    "CostFunctionJb",
    "CostFunctionJo",
    "GradientOfCostFunctionJ",
    "GradientOfCostFunctionJb",
    "GradientOfCostFunctionJo",
    "SigmaObs2",
    "SigmaBck2",
    "APosterioriCorrelations",
    "APosterioriCovariance",
    "APosterioriStandardDeviations",
    "APosterioriVariances",
    "Residu",
    ]

# Global regulation of separate container execution with priority on user:
#   0 for no separate execution container possibility
#   1 for separate execution container possibility
UseSeparateContainer = 1



# -*- coding: utf-8 -*-
#
# Copyright (C) 2008-2022 EDF R&D
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
#
# See http://www.salome-platform.org/ or email : webmaster.salome@opencascade.com
#
# Author: Jean-Philippe Argaud, jean-philippe.argaud@edf.fr, EDF R&D

import numpy
from daCore import BasicObjects

# ==============================================================================
class ElementaryAlgorithm(BasicObjects.Algorithm):
    def __init__(self):
        BasicObjects.Algorithm.__init__(self, "INPUTVALUESTEST")
        self.defineRequiredParameter(
            name     = "NumberOfPrintedDigits",
            default  = 5,
            typecast = int,
            message  = "Nombre de chiffres affichés pour les impressions de réels",
            minval   = 0,
            )
        self.defineRequiredParameter(
            name     = "PrintAllValuesFor",
            default  = [],
            typecast = tuple,
            message  = "Liste de noms de vecteurs dont les valeurs détaillées sont à imprimer",
            listval  = [
                "Background",
                "CheckingPoint",
                "Observation",
                ]
            )
        self.defineRequiredParameter(
            name     = "ShowInformationOnlyFor",
            default  = ["Background", "CheckingPoint", "Observation"],
            typecast = tuple,
            message  = "Liste de noms de vecteurs dont les informations synthétiques sont à imprimer",
            listval  = [
                "Background",
                "CheckingPoint",
                "Observation",
                ]
            )
        self.defineRequiredParameter(
            name     = "SetDebug",
            default  = False,
            typecast = bool,
            message  = "Activation du mode debug lors de l'exécution",
            )
        self.requireInputArguments(
            mandatory= (),
            )
        self.setAttributes(tags=(
            "Checking",
            ))

    def run(self, Xb=None, Y=None, U=None, HO=None, EM=None, CM=None, R=None, B=None, Q=None, Parameters=None):
        self._pre_run(Parameters, Xb, Y, U, HO, EM, CM, R, B, Q)
        #
        _p = self._parameters["NumberOfPrintedDigits"]
        numpy.set_printoptions(precision=_p)
        #
        def __buildPrintableVectorProperties( __name, __vector ):
            if __vector is None:                                         return ""
            if len(__vector) == 0:                                       return ""
            if hasattr(__vector,"name") and __name != __vector.name():   return ""
            if __name not in self._parameters["ShowInformationOnlyFor"]: return ""
            #
            if hasattr(__vector,"mins"):
                __title = "Information for %svector series:"%(str(__name)+" ",)
            else:
                __title = "Information for %svector:"%(str(__name)+" ",)
            msgs = "\n"
            msgs += ("===> "+__title+"\n")
            msgs += ("     "+("-"*len(__title))+"\n")
            msgs += ("     Main characteristics of the vector:\n")
            if hasattr(__vector,"basetype"):
                msgs += ("       Python base type..........: %s\n")%( __vector.basetype(), )
                msgs += ("       Shape of data.............: %s\n")%( __vector.shape(), )
            else:
                msgs += ("       Python base type..........: %s\n")%type( __vector )
                msgs += ("       Shape of serie of vectors.: %s\n")%( __vector.shape, )
            try:
                msgs += ("       Number of data............: %s\n")%( len(__vector), )
            except: pass
            if hasattr(__vector,"mins"):
                msgs += ("       Serie of minimum values...: %s\n")%numpy.array(__vector.mins())
            else:
                msgs += ("       Minimum of vector.........: %12."+str(_p)+"e\n")%__vector.min()
            if hasattr(__vector,"means"):
                msgs += ("       Serie of mean values......: %s\n")%numpy.array(__vector.means())
            else:
                msgs += ("       Mean of vector............: %12."+str(_p)+"e\n")%__vector.mean()
            if hasattr(__vector,"maxs"):
                msgs += ("       Serie of maximum values...: %s\n")%numpy.array(__vector.maxs())
            else:
                msgs += ("       Maximum of vector.........: %12."+str(_p)+"e\n")%__vector.max()
            if self._parameters["SetDebug"] or __name in self._parameters["PrintAllValuesFor"]:
                msgs += ("\n")
                msgs += ("     Printing all values :\n")
                msgs += ("%s"%(__vector,))
            print(msgs)
            return msgs
        #----------
        __buildPrintableVectorProperties( "Background",    Xb )
        __buildPrintableVectorProperties( "CheckingPoint", Xb )
        __buildPrintableVectorProperties( "Observation",    Y )
        #
        self._post_run(HO)
        return 0

# ==============================================================================
if __name__ == "__main__":
    print('\n AUTODIAGNOSTIC\n')
